__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import Expression, Type


class Boolean(Expression):
    """
    Convert to boolean truth value.

    @ivar expr: (Integer) expression to convert.
    @type expr: C{Expression}
    """

    def __init__(self, expr, pos=None):
        Expression.__init__(self, pos)
        self.expr = expr

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Force expression to boolean:")
        self.expr.debug_print(indentation + 2)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        expr = self.expr.reduce(id_dicts)
        if expr.type() != Type.INTEGER:
            if expr.type() == Type.SPRITEGROUP_REF:
                raise generic.ProcCallSyntaxError(expr.name, expr.pos)
            raise generic.ScriptError("Only integers can be converted to a boolean value.", expr.pos)
        if expr.is_boolean():
            return expr
        return Boolean(expr)

    def supported_by_action2(self, raise_error):
        return self.expr.supported_by_action2(raise_error)

    def supported_by_actionD(self, raise_error):
        return self.expr.supported_by_actionD(raise_error)

    def collect_references(self):
        return self.expr.collect_references()

    def is_read_only(self):
        return self.expr.is_read_only()

    def is_boolean(self):
        return True

    def __str__(self):
        return "!!({})".format(self.expr)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import re

from .array import Array
from .base_expression import ConstantFloat, ConstantNumeric, Expression, Type
from .bin_not import BinNot, Not
from .binop import BinOp
from .bitmask import BitMask
from .boolean import Boolean
from .cargo import AcceptCargo, ProduceCargo
from .functioncall import FunctionCall, GRMOp, SpecialCheck
from .functionptr import FunctionPtr
from .identifier import Identifier
from .parameter import OtherGRFParameter, Parameter, parse_string_to_dword
from .patch_variable import PatchVariable
from .special_parameter import SpecialParameter
from .spritegroup_ref import SpriteGroupRef
from .storage_op import StorageOp
from .string import String
from .string_literal import StringLiteral
from .ternaryop import TernaryOp
from .abs_op import AbsOp
from .variable import Variable

is_valid_id = re.compile("[a-zA-Z_][a-zA-Z0-9_]{3}$")


def identifier_to_print(name):
    """
    Check whether the given name is a valid 4 letter identifier to print
    (for cargoes, railtypes, roadtypes and tramtypes).

    @param name: Name to check.
    @return The identifier itself, if it is a valid name, else a string literal text with the name.
    """
    if is_valid_id.match(name):
        return name
    return '"{}"'.format(name)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action0properties
from nml.ast import assignment, base_statement


class TileLayout(base_statement.BaseStatement):
    """
    'tile_layout' AST node. A TileLayout is a list of x,y-offset/tileID pairs.
    The x and y offsets are from the northernmost tile of the industry/airport.
    Additionally, some extra properties can be stored in the TileLayout, like
    the orientation of the airport.

    @ivar name: The name of this layout by which it can be referenced later.
    @type name: C{str}

    @ivar tile_prop_list: List of offset/tileID and properties.
    @type tile_prop_list: C{list} of L{LayoutTile} and L{Assignment}

    @ivar tile_list: List of tile-offsets/tileIDs.
    @type tile_list: C{list} of C{LayoutTile} with constant x and y values.

    @ivar properties: table of all properties. Unknown property names are accepted and ignored.
    @type properties: C{dict} with C{str} keys and L{ConstantNumeric} values
    """

    def __init__(self, name, tile_list, pos):
        base_statement.BaseStatement.__init__(self, "tile layout", pos, False, False)
        self.name = name.value
        self.tile_prop_list = tile_list
        self.tile_list = []
        self.properties = {}

    def pre_process(self):
        for tileprop in self.tile_prop_list:
            if isinstance(tileprop, assignment.Assignment):
                name = tileprop.name.value
                if name in self.properties:
                    raise generic.ScriptError("Duplicate property {} in tile layout".format(name), tileprop.name.pos)
                self.properties[name] = tileprop.value.reduce_constant(global_constants.const_list)
            else:
                assert isinstance(tileprop, LayoutTile)
                x = tileprop.x.reduce_constant()
                y = tileprop.y.reduce_constant()
                tile = tileprop.tiletype.reduce(unknown_id_fatal=False)
                if isinstance(tile, expression.Identifier) and tile.value == "clear":
                    tile = expression.ConstantNumeric(0xFF)
                self.tile_list.append(LayoutTile(x, y, tile))
        if self.name in action0properties.tilelayout_names:
            raise generic.ScriptError(
                "A tile layout with name '{}' has already been defined.".format(self.name), self.pos
            )
        action0properties.tilelayout_names[self.name] = self

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "TileLayout")
        for tile in self.tile_list:
            generic.print_dbg(indentation + 2, "At {:d},{:d}:".format(tile.x, tile.y))
            tile.tiletype.debug_print(indentation + 4)

    def get_action_list(self):
        return []

    def __str__(self):
        return "tilelayout {} {{\n\t{}\n}}\n".format(self.name, "\n\t".join(str(x) for x in self.tile_prop_list))

    def get_size(self):
        size = 2
        for tile in self.tile_list:
            size += 3
            if not isinstance(tile.tiletype, expression.ConstantNumeric):
                size += 2
        return size

    def write(self, file):
        for tile in self.tile_list:
            file.print_bytex(tile.x.value)
            file.print_bytex(tile.y.value)
            if isinstance(tile.tiletype, expression.ConstantNumeric):
                file.print_bytex(tile.tiletype.value)
            else:
                if not isinstance(tile.tiletype, expression.Identifier):
                    raise generic.ScriptError("Invalid expression type for layout tile", tile.tiletype.pos)
                if tile.tiletype.value not in global_constants.item_names:
                    raise generic.ScriptError("Unknown tile name", tile.tiletype.pos)
                file.print_bytex(0xFE)
                tile_id = global_constants.item_names[tile.tiletype.value].id
                if not isinstance(tile_id, expression.ConstantNumeric):
                    raise generic.ScriptError(
                        "Tile '{}' cannot be used in a tilelayout, as its ID is not a constant.".format(
                            tile.tiletype.value
                        ),
                        tile.tiletype.pos,
                    )
                file.print_wordx(tile_id.value)
            file.newline()
        file.print_bytex(0)
        file.print_bytex(0x80)
        file.newline()


class LayoutTile:
    """
    Single tile that is part of a L{TileLayout}.

    @ivar x: X-offset from the northernmost tile of the industry/airport.
    @type x: L{Expression}

    @ivar y: Y-offset from the northernmost tile of the industry/airport.
    @type y: L{Expression}

    @ivar tiletype: TileID of the tile to draw on the given offset.
    @type tiletype: L{Expression}
    """

    def __init__(self, x, y, tiletype):
        self.x = x
        self.y = y
        self.tiletype = tiletype

    def __str__(self):
        return "{}, {}: {};".format(self.x, self.y, self.tiletype)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, nmlop

def default_60xvar(name, args, pos, info):
    """
    Function to convert arguments into a variable parameter.
    This function handles the default case of one argument.

    @param name: Name of the variable
    @type name: C{str}

    @param args: List of passed arguments
    @type args: C{list} of L{Expression}

    @param pos: Position information
    @type pos: L{Position}

    @param info: Information of the variable, as found in the dictionary
    @type info: C{dict}

    @return: A tuple of two values:
                - Parameter to use for the 60+x variable
                - List of possible extra parameters that need to be passed via registers
    @rtype: C{tuple} of (L{Expression}, C{list} C{tuple} of (C{int}, L{Expression}))
    """
    if len(args) != 1:
        raise generic.ScriptError("'{}'() requires one argument, encountered {:d}".format(name, len(args)), pos)
    return (args[0], [])

# Some commonly used functions that apply some modification to the raw variable value
# To pass extra parameters, lambda calculus may be used

def value_sign_extend(var, info):
    #r = (x ^ m) - m; with m being (1 << (num_bits -1))
    m = expression.ConstantNumeric(1 << (info['size'] - 1))
    return nmlop.SUB(nmlop.XOR(var, m), m)

def value_mul_div(mul, div):
    return lambda var, info: nmlop.DIV(nmlop.MUL(var, mul), div)

def value_add_constant(const):
    return lambda var, info: nmlop.ADD(var, const)

def value_equals(const):
    return lambda var, info: nmlop.CMP_EQ(var, const)

# Commonly used functions to let a variable accept an (x, y)-offset as parameters

def tile_offset(name, args, pos, info, min, max):
    if len(args) != 2:
        raise generic.ScriptError("'{}'() requires 2 arguments, encountered {:d}".format(name, len(args)), pos)
    for arg in args:
        if isinstance(arg, expression.ConstantNumeric):
            generic.check_range(arg.value, min, max, "Argument of '{}'".format(name), arg.pos)

    x = nmlop.AND(args[0], 0xF)
    y = nmlop.AND(args[1], 0xF)
    # Shift y left by four
    y = nmlop.SHIFT_LEFT(y, 4)
    param = nmlop.ADD(x, y)
    #Make sure to reduce the result
    return ( param.reduce(), [] )

def signed_tile_offset(name, args, pos, info):
    return tile_offset(name, args, pos, info, -8, 7)

def unsigned_tile_offset(name, args, pos, info):
    return tile_offset(name, args, pos, info, 0, 15)

#
# Global variables, usable for all features
#

varact2_globalvars = {
    'current_month'        : {'var': 0x02, 'start':  0, 'size':  8},
    'current_day_of_month' : {'var': 0x02, 'start':  8, 'size':  5},
    'is_leapyear'          : {'var': 0x02, 'start': 15, 'size':  1},
    'current_day_of_year'  : {'var': 0x02, 'start': 16, 'size':  9},
    'traffic_side'         : {'var': 0x06, 'start':  4, 'size':  1},
    'animation_counter'    : {'var': 0x0A, 'start':  0, 'size': 16},
    'current_callback'     : {'var': 0x0C, 'start':  0, 'size': 16},
    'extra_callback_info1' : {'var': 0x10, 'start':  0, 'size': 32},
    'game_mode'            : {'var': 0x12, 'start':  0, 'size':  8},
    'extra_callback_info2' : {'var': 0x18, 'start':  0, 'size': 32},
    'display_options'      : {'var': 0x1B, 'start':  0, 'size':  6},
    'last_computed_result' : {'var': 0x1C, 'start':  0, 'size': 32},
    'snowline_height'      : {'var': 0x20, 'start':  0, 'size':  8},
    'difficulty_level'     : {'var': 0x22, 'start':  0, 'size':  8},
    'current_date'         : {'var': 0x23, 'start':  0, 'size': 32},
    'current_year'         : {'var': 0x24, 'start':  0, 'size': 32},
}

#
# Vehicles (features 0x00 - 0x03)
# A few variables have an implementation that differs per vehicle type
# These are to be found below
#

varact2vars_vehicles = {
    'grfid'                            : {'var': 0x25, 'start':  0, 'size': 32},
    'position_in_consist'              : {'var': 0x40, 'start':  0, 'size':  8},
    'position_in_consist_from_end'     : {'var': 0x40, 'start':  8, 'size':  8},
    'num_vehs_in_consist'              : {'var': 0x40, 'start': 16, 'size':  8, 'value_function': value_add_constant(1)}, # Zero-based, add 1 to make sane
    'position_in_vehid_chain'          : {'var': 0x41, 'start':  0, 'size':  8},
    'position_in_vehid_chain_from_end' : {'var': 0x41, 'start':  8, 'size':  8},
    'num_vehs_in_vehid_chain'          : {'var': 0x41, 'start': 16, 'size':  8}, # One-based, already sane
    'cargo_classes_in_consist'         : {'var': 0x42, 'start':  0, 'size':  8},
    'most_common_cargo_type'           : {'var': 0x42, 'start':  8, 'size':  8},
    'most_common_cargo_subtype'        : {'var': 0x42, 'start': 16, 'size':  8},
    'bitmask_consist_info'             : {'var': 0x42, 'start': 24, 'size':  8},
    'company_num'                      : {'var': 0x43, 'start':  0, 'size':  8},
    'company_type'                     : {'var': 0x43, 'start': 16, 'size':  2},
    'company_colour1'                  : {'var': 0x43, 'start': 24, 'size':  4},
    'company_colour2'                  : {'var': 0x43, 'start': 28, 'size':  4},
    'aircraft_height'                  : {'var': 0x44, 'start':  8, 'size':  8},
    'airport_type'                     : {'var': 0x44, 'start':  0, 'size':  8},
    'curv_info_prev_cur'               : {'var': 0x45, 'start':  0, 'size':  4, 'value_function': value_sign_extend},
    'curv_info_cur_next'               : {'var': 0x45, 'start':  8, 'size':  4, 'value_function': value_sign_extend},
    'curv_info_prev_next'              : {'var': 0x45, 'start': 16, 'size':  4, 'value_function': value_sign_extend},
    'curv_info'                        : {'var': 0x45, 'start':  0, 'size': 12,
            'value_function': lambda var, info: nmlop.AND(var, 0x0F0F).reduce()},
    'motion_counter'                   : {'var': 0x46, 'start':  8, 'size': 24},
    'cargo_type_in_veh'                : {'var': 0x47, 'start':  0, 'size':  8},
    'cargo_unit_weight'                : {'var': 0x47, 'start':  8, 'size':  8},
    'cargo_classes'                    : {'var': 0x47, 'start': 16, 'size': 16},
    'vehicle_is_available'             : {'var': 0x48, 'start':  0, 'size':  1},
    'vehicle_is_testing'               : {'var': 0x48, 'start':  1, 'size':  1},
    'vehicle_is_offered'               : {'var': 0x48, 'start':  2, 'size':  1},
    'build_year'                       : {'var': 0x49, 'start':  0, 'size': 32},
    'vehicle_is_potentially_powered'   : {'var': 0x4A, 'start':  8, 'size':  1},
    'tile_has_catenary'                : {'var': 0x4A, 'start':  9, 'size':  1},
    'date_of_last_service'             : {'var': 0x4B, 'start':  0, 'size': 32},
    'position_in_articulated_veh'          : {'var': 0x4D, 'start':  0, 'size':  8},
    'position_in_articulated_veh_from_end' : {'var': 0x4D, 'start':  8, 'size':  8},
    'waiting_triggers'                 : {'var': 0x5F, 'start':  0, 'size':  8},
    'random_bits'                      : {'var': 0x5F, 'start':  8, 'size': 16},
    'direction'                        : {'var': 0x9F, 'start':  0, 'size':  8},
    'vehicle_is_hidden'                : {'var': 0xB2, 'start':  0, 'size':  1},
    'vehicle_is_stopped'               : {'var': 0xB2, 'start':  1, 'size':  1},
    'vehicle_is_crashed'               : {'var': 0xB2, 'start':  7, 'size':  1},
    'cargo_capacity'                   : {'var': 0xBA, 'start':  0, 'size': 16},
    'cargo_count'                      : {'var': 0xBC, 'start':  0, 'size': 16},
    'age_in_days'                      : {'var': 0xC0, 'start':  0, 'size': 16},
    'max_age_in_days'                  : {'var': 0xC2, 'start':  0, 'size': 16},
    'vehicle_type_id'                  : {'var': 0xC6, 'start':  0, 'size': 16},
    'vehicle_is_flipped'               : {'var': 0xC8, 'start':  1, 'size':  1},
    'breakdowns_since_last_service'    : {'var': 0xCA, 'start':  0, 'size':  8},
    'vehicle_is_broken'                : {'var': 0xCB, 'start':  0, 'size':  8, 'value_function': value_equals(1)},
    'reliability'                      : {'var': 0xCE, 'start':  0, 'size': 16, 'value_function': value_mul_div(101, 0x10000)},
    'cargo_subtype'                    : {'var': 0xF2, 'start':  0, 'size':  8},
    'vehicle_is_unloading'             : {'var': 0xFE, 'start':  1, 'size':  1},
    'vehicle_is_powered'               : {'var': 0xFE, 'start':  5, 'size':  1},
    'vehicle_is_not_powered'           : {'var': 0xFE, 'start':  6, 'size':  1},
    'vehicle_is_reversed'              : {'var': 0xFE, 'start':  8, 'size':  1},
    'built_during_preview'             : {'var': 0xFE, 'start': 10, 'size':  1},
}

# Vehicle-type-specific variables

varact2vars_trains = {
    **varact2vars_vehicles,
    #0x4786 / 0x10000 is an approximation of 3.5790976, the conversion factor
    #for train speed
    'max_speed'           : {'var': 0x98, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x4786, 0x10000)},
    'current_speed'       : {'var': 0xB4, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x4786, 0x10000)},
    'current_railtype'    : {'var': 0x4A, 'start':  0, 'size':  8},
    'current_max_speed'   : {'var': 0x4C, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x4786, 0x10000)},
    'vehicle_is_in_depot' : {'var': 0xE2, 'start': 7, 'size':  1},
}

varact2vars_roadvehs = {
    **varact2vars_vehicles,
    #0x23C3 / 0x10000 is an approximation of 7.1581952, the conversion factor
    #for road vehicle speed
    'max_speed'           : {'var': 0x98, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x23C3, 0x10000)},
    'current_speed'       : {'var': 0xB4, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x23C3, 0x10000)},
    'current_roadtype'    : {'var': 0x4A, 'start':  0, 'size':  8},
    'current_tramtype'    : {'var': 0x4A, 'start':  0, 'size':  8},
    'current_max_speed'   : {'var': 0x4C, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x23C3, 0x10000)},
    'vehicle_is_in_depot' : {'var': 0xE2, 'start': 0, 'size':  8, 'value_function': value_equals(0xFE)},
}

varact2vars_ships = {
    **varact2vars_vehicles,
    #0x23C3 / 0x10000 is an approximation of 7.1581952, the conversion factor
    #for ship speed
    'max_speed'           : {'var': 0x98, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x23C3, 0x10000)},
    'current_speed'       : {'var': 0xB4, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x23C3, 0x10000)},
    'current_max_speed'   : {'var': 0x4C, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x23C3, 0x10000)},
    'vehicle_is_in_depot' : {'var': 0xE2, 'start': 7, 'size':  1},
}

varact2vars_aircraft = {
    **varact2vars_vehicles,
    #0x3939 / 0x1000 is an approximation of 0.279617, the conversion factor
    #Note that the denominator has one less zero here!
    #for aircraft speed
    'max_speed'           : {'var': 0x98, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x3939, 0x1000)},
    'current_speed'       : {'var': 0xB4, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x3939, 0x1000)},
    'current_max_speed'   : {'var': 0x4C, 'start': 0, 'size': 16, 'value_function': value_mul_div(0x3939, 0x1000)},
    'flight_state'        : {'var': 0xE2, 'start': 0, 'size':  8},
    'vehicle_is_in_depot' : {'var': 0xE6, 'start': 0, 'size':  8, 'value_function': value_equals(0)},
}

def signed_byte_parameter(name, args, pos, info):
    # Convert to a signed byte by AND-ing with 0xFF
    if len(args) != 1:
        raise generic.ScriptError("{}() requires one argument, encountered {:d}".format(name, len(args)), pos)
    if isinstance(args[0], expression.ConstantNumeric):

        generic.check_range(args[0].value, -128, 127, "parameter of {}()".format(name), pos)
    ret = nmlop.AND(args[0], 0xFF, pos).reduce()
    return (ret, [])

def vehicle_railtype(name, args, pos, info):
    return (expression.functioncall.builtin_resolve_typelabel(name, args, pos, table_name="railtype"), [])

def vehicle_roadtype(name, args, pos, info):
    return (expression.functioncall.builtin_resolve_typelabel(name, args, pos, table_name="roadtype"), [])

def vehicle_tramtype(name, args, pos, info):
    return (expression.functioncall.builtin_resolve_typelabel(name, args, pos, table_name="tramtype"), [])

varact2vars60x_vehicles = {
    'count_veh_id'        : {'var': 0x60, 'start':  0, 'size': 8},
    'other_veh_curv_info' : {'var': 0x62, 'start':  0, 'size': 4, 'param_function':signed_byte_parameter, 'value_function':value_sign_extend},
    'other_veh_is_hidden' : {'var': 0x62, 'start':  7, 'size': 1, 'param_function':signed_byte_parameter},
    'other_veh_x_offset'  : {'var': 0x62, 'start':  8, 'size': 8, 'param_function':signed_byte_parameter, 'value_function':value_sign_extend},
    'other_veh_y_offset'  : {'var': 0x62, 'start': 16, 'size': 8, 'param_function':signed_byte_parameter, 'value_function':value_sign_extend},
    'other_veh_z_offset'  : {'var': 0x62, 'start': 24, 'size': 8, 'param_function':signed_byte_parameter, 'value_function':value_sign_extend},
}

varact2vars60x_trains = {
    **varact2vars60x_vehicles,
    # Var 0x63 bit 0 is only useful when testing multiple bits at once. On its own it is already covered by railtype_available().
    'tile_supports_railtype' : {'var': 0x63, 'start':  1, 'size': 1, 'param_function':vehicle_railtype},
    'tile_powers_railtype'   : {'var': 0x63, 'start':  2, 'size': 1, 'param_function':vehicle_railtype},
    'tile_is_railtype'       : {'var': 0x63, 'start':  3, 'size': 1, 'param_function':vehicle_railtype},
}

varact2vars60x_roadvehs = {
    **varact2vars60x_vehicles,
    # Var 0x63 bit 0 is only useful when testing multiple bits at once. On its own it is already covered by road/tramtype_available().
    'tile_supports_roadtype' : {'var': 0x63, 'start':  1, 'size': 1, 'param_function':vehicle_roadtype},
    'tile_supports_tramtype' : {'var': 0x63, 'start':  1, 'size': 1, 'param_function':vehicle_tramtype},
    'tile_powers_roadtype'   : {'var': 0x63, 'start':  2, 'size': 1, 'param_function':vehicle_roadtype},
    'tile_powers_tramtype'   : {'var': 0x63, 'start':  2, 'size': 1, 'param_function':vehicle_tramtype},
    'tile_is_roadtype'       : {'var': 0x63, 'start':  3, 'size': 1, 'param_function':vehicle_roadtype},
    'tile_is_tramtype'       : {'var': 0x63, 'start':  3, 'size': 1, 'param_function':vehicle_tramtype},
}

#
# Stations (feature 0x04)
#

# 'Base station' variables are shared between stations and airports
varact2vars_base_stations = {
    'random_bits_station' : {'var': 0x5F, 'start': 8, 'size': 16},
    # Var 48 doesn't work with newcargos, do not use
    'had_vehicle_of_type' : {'var': 0x8A, 'start': 1, 'size': 5}, # Only read bits 1-5
    'is_waypoint'         : {'var': 0x8A, 'start': 6, 'size': 1},
    'facilities'          : {'var': 0xF0, 'start': 0, 'size': 8},
    'airport_type'        : {'var': 0xF1, 'start': 0, 'size': 8},
    # Variables F2, F3, F6 (roadstop, airport flags) are next to useless
    # Also, their values are not the same as in TTDP / spec
    # Therefore, these are not implemented
    'build_date'          : {'var': 0xFA, 'start': 0, 'size': 16, 'value_function': value_add_constant(701265)}
}

varact2vars60x_base_stations = {
    'cargo_amount_waiting'      : {'var': 0x60, 'start': 0, 'size': 32},
    'cargo_time_since_pickup'   : {'var': 0x61, 'start': 0, 'size': 32},
    'cargo_rating'              : {'var': 0x62, 'start': 0, 'size': 32, 'value_function': value_mul_div(101, 256)},
    'cargo_time_en_route'       : {'var': 0x63, 'start': 0, 'size': 32},
    'cargo_last_vehicle_speed'  : {'var': 0x64, 'start': 0, 'size':  8},
    'cargo_last_vehicle_age'    : {'var': 0x64, 'start': 8, 'size':  8},
    'cargo_accepted'            : {'var': 0x65, 'start': 3, 'size':  1},
    'cargo_accepted_ever'       : {'var': 0x69, 'start': 0, 'size':  1},
    'cargo_accepted_last_month' : {'var': 0x69, 'start': 1, 'size':  1},
    'cargo_accepted_this_month' : {'var': 0x69, 'start': 2, 'size':  1},
    'cargo_accepted_bigtick'    : {'var': 0x69, 'start': 3, 'size':  1},
}

varact2vars_stations = {
    **varact2vars_base_stations,
    # Vars 40, 41, 46, 47, 49 are implemented as 60+x vars,
    # except for the 'tile type' part which is always the same anyway
    'tile_type'                : {'var': 0x40, 'start': 24, 'size': 8},
    'terrain_type'             : {'var': 0x42, 'start':  0, 'size': 8},
    'track_type'               : {'var': 0x42, 'start':  8, 'size': 8},
    'company_num'              : {'var': 0x43, 'start':  0, 'size': 8},
    'company_type'             : {'var': 0x43, 'start': 16, 'size': 2},
    'company_colour1'          : {'var': 0x43, 'start': 24, 'size': 4},
    'company_colour2'          : {'var': 0x43, 'start': 28, 'size': 4},
    'pbs_reserved'             : {'var': 0x44, 'start':  0, 'size': 1},
    'pbs_reserved_or_disabled' : {'var': 0x44, 'start':  1, 'size': 1},
    'pbs_enabled'              : {'var': 0x44, 'start':  2, 'size': 1},
    'rail_continuation'        : {'var': 0x45, 'start':  0, 'size': 8},
    'rail_present'             : {'var': 0x45, 'start':  8, 'size': 8},
    'animation_frame'          : {'var': 0x4A, 'start':  0, 'size': 8},
    'random_bits_tile'         : {'var': 0x5F, 'start': 24, 'size': 4},
}

# Mapping of param values for platform_xx vars to variable numbers
mapping_platform_param = {
    (0, False) : 0x40,
    (1, False) : 0x41,
    (0, True)  : 0x46,
    (1, True)  : 0x47,
    (2, False) : 0x49,
}

def platform_info_param(name, args, pos, info):
    if len(args) != 1:
        raise generic.ScriptError("'{}'() requires one argument, encountered {:d}".format(name, len(args)), pos)
    if (not isinstance(args[0], expression.ConstantNumeric)) or args[0].value not in (0, 1, 2):
        raise generic.ScriptError("Invalid argument for '{}'(), must be one of PLATFORM_SAME_XXX.".format(name), pos)

    is_middle = 'middle' in info and info['middle']
    if is_middle and args[0].value == 2:
        raise generic.ScriptError("Invalid argument for '{}'(), PLATFORM_SAME_DIRECTION is not supported here.".format(name), pos)
    # Temporarily store variable number in the param, this will be fixed in the value_function
    return (expression.ConstantNumeric(mapping_platform_param[(args[0].value, is_middle)]), [])

def platform_info_fix_var(var, info):
    # Variable to use was temporarily stored in the param
    # Fix this now
    var.num = var.param
    var.param = None
    return var

varact2vars60x_stations = {
    **varact2vars60x_base_stations,
    'nearby_tile_animation_frame'   : {'var': 0x66, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
    'nearby_tile_slope'             : {'var': 0x67, 'start':  0, 'size':  5, 'param_function': signed_tile_offset},
    'nearby_tile_is_water'          : {'var': 0x67, 'start':  9, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_terrain_type'      : {'var': 0x67, 'start': 10, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_water_class'       : {'var': 0x67, 'start': 13, 'size':  2, 'param_function': signed_tile_offset},
    'nearby_tile_height'            : {'var': 0x67, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_class'             : {'var': 0x67, 'start': 24, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_is_station'        : {'var': 0x68, 'start':  0, 'size': 32, 'param_function': signed_tile_offset, 'value_function': lambda var, info: nmlop.CMP_NEQ(var, 0xFFFFFFFF)},
    'nearby_tile_original_gfx'      : {'var': 0x68, 'start':  8, 'size':  2, 'param_function': signed_tile_offset, 'value_function': value_equals(0)},
    'nearby_tile_same_grf'          : {'var': 0x68, 'start':  8, 'size':  2, 'param_function': signed_tile_offset, 'value_function': value_equals(1)},
    'nearby_tile_other_grf'         : {'var': 0x68, 'start':  8, 'size':  2, 'param_function': signed_tile_offset, 'value_function': value_equals(2)},
    'nearby_tile_same_station'      : {'var': 0x68, 'start': 10, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_perpendicular'     : {'var': 0x68, 'start': 11, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_tile_type'         : {'var': 0x68, 'start': 11, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_grfid'             : {'var': 0x6A, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
    'nearby_tile_station_id'        : {'var': 0x6B, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
    # 'var' will be set in the value_function, depending on parameter
    'platform_length'               : {'var': 0x00, 'start': 16, 'size':  4, 'param_function': platform_info_param, 'value_function': platform_info_fix_var},
    'platform_count'                : {'var': 0x00, 'start': 20, 'size':  4, 'param_function': platform_info_param, 'value_function': platform_info_fix_var},
    'platform_position_from_start'  : {'var': 0x00, 'start':  0, 'size':  4, 'param_function': platform_info_param, 'value_function': platform_info_fix_var},
    'platform_position_from_end'    : {'var': 0x00, 'start':  4, 'size':  4, 'param_function': platform_info_param, 'value_function': platform_info_fix_var},
    'platform_number_from_start'    : {'var': 0x00, 'start':  8, 'size':  4, 'param_function': platform_info_param, 'value_function': platform_info_fix_var},
    'platform_number_from_end'      : {'var': 0x00, 'start': 12, 'size':  4, 'param_function': platform_info_param, 'value_function': platform_info_fix_var},
    'platform_position_from_middle' : {'var': 0x00, 'start':  0, 'size':  4, 'param_function': platform_info_param, 'middle': True, # 'middle' is used by platform_info_param
                                            'value_function': lambda var, info: value_sign_extend(platform_info_fix_var(var, info), info)},
    'platform_number_from_middle'   : {'var': 0x00, 'start':  4, 'size':  4, 'param_function': platform_info_param, 'middle': True, # 'middle' is used by platform_info_param
                                            'value_function': lambda var, info: value_sign_extend(platform_info_fix_var(var, info), info)},
}

#
# Canals (feature 0x05)
#

varact2vars_canals = {
    'tile_height'  : {'var': 0x80, 'start': 0, 'size': 8},
    'terrain_type' : {'var': 0x81, 'start': 0, 'size': 8},
    'dike_map'     : {'var': 0x82, 'start': 0, 'size': 8},
    'random_bits'  : {'var': 0x83, 'start': 0, 'size': 8},
}
# Canals have no 60+X variables

#
# Bridges (feature 0x06) have no variational action2
#

#
# Houses (feature 0x07)
#

def house_same_class(var, info):
    # Just using var 44 fails for non-north house tiles, as these have no class
    # Therefore work around it using var 61
    # Load ID of the north tile from register FF bits 24..31, and use that as param for var 61
    north_tile = expression.Variable(expression.ConstantNumeric(0x7D), expression.ConstantNumeric(24),
                                     expression.ConstantNumeric(0xFF), expression.ConstantNumeric(0xFF), var.pos)
    var61 = expression.Variable(expression.ConstantNumeric(0x7B), expression.ConstantNumeric(info['start']),
                                     expression.ConstantNumeric((1 << info['size']) - 1), expression.ConstantNumeric(0x61), var.pos)
    return nmlop.VAL2(north_tile, var61, var.pos)


varact2vars_houses = {
    'construction_state'    : {'var': 0x40, 'start':  0, 'size':  2},
    'pseudo_random_bits'    : {'var': 0x40, 'start':  2, 'size':  2},
    'age'                   : {'var': 0x41, 'start':  0, 'size':  8},
    'town_zone'             : {'var': 0x42, 'start':  0, 'size':  8},
    'terrain_type'          : {'var': 0x43, 'start':  0, 'size':  8},
    'same_house_count_town' : {'var': 0x44, 'start':  0, 'size':  8},
    'same_house_count_map'  : {'var': 0x44, 'start':  8, 'size':  8},
    'same_class_count_town' : {'var': 0xFF, 'start': 16, 'size':  8, 'value_function': house_same_class}, # 'var' is unused
    'same_class_count_map'  : {'var': 0xFF, 'start': 24, 'size':  8, 'value_function': house_same_class}, # 'var' is unused
    'generating_town'       : {'var': 0x45, 'start':  0, 'size':  1},
    'animation_frame'       : {'var': 0x46, 'start':  0, 'size':  8},
    'x_coordinate'          : {'var': 0x47, 'start':  0, 'size': 16},
    'y_coordinate'          : {'var': 0x47, 'start': 16, 'size': 16},
    'random_bits'           : {'var': 0x5F, 'start':  8, 'size':  8},
    'relative_x'            : {'var': 0x7D, 'start':  0, 'size':  8, 'param': 0xFF},
    'relative_y'            : {'var': 0x7D, 'start':  8, 'size':  8, 'param': 0xFF},
    'relative_pos'          : {'var': 0x7D, 'start':  0, 'size': 16, 'param': 0xFF},
    'house_tile'            : {'var': 0x7D, 'start': 16, 'size':  8, 'param': 0xFF},
    'house_type_id'         : {'var': 0x7D, 'start': 24, 'size':  8, 'param': 0xFF},
}

def cargo_accepted_nearby(name, args, pos, info):
    # cargo_accepted_nearby(cargo[, xoffset, yoffset])
    if len(args) not in (1, 3):
        raise generic.ScriptError("{}() requires 1 or 3 arguments, encountered {:d}".format(name, len(args)), pos)

    if len(args) > 1:
        offsets = args[1:3]
        for i, offs in enumerate(offsets[:]):
            if isinstance(offs, expression.ConstantNumeric):
                generic.check_range(offs.value, -128, 127, "{}-parameter {:d} '{}offset'".format(name, i + 1, "x" if i == 0 else "y"), pos)
            offsets[i] = nmlop.AND(offs, 0xFF, pos).reduce()
        # Register 0x100 should be set to xoffset | (yoffset << 8)
        reg100 = nmlop.OR(nmlop.MUL(offsets[1], 256, pos), offsets[0]).reduce()
    else:
        reg100 = expression.ConstantNumeric(0, pos)

    return (args[0], [(0x100, reg100)])

def nearest_house_matching_criterion(name, args, pos, info):
    # nearest_house_matching_criterion(radius, criterion)
    # parameter is radius | (criterion << 6)
    if len(args) != 2:
        raise generic.ScriptError("{}() requires 2 arguments, encountered {:d}".format(name, len(args)), pos)
    if isinstance(args[0], expression.ConstantNumeric):
        generic.check_range(args[0].value, 1, 63, "{}()-parameter 1 'radius'".format(name), pos)
    if isinstance(args[1], expression.ConstantNumeric) and args[1].value not in (0, 1, 2):
        raise generic.ScriptError("Invalid value for {}()-parameter 2 'criterion'".format(name), pos)

    radius = nmlop.AND(args[0], 0x3F, pos)
    criterion = nmlop.AND(args[1], 0x03, pos)
    criterion = nmlop.MUL(criterion, 0x40)
    retval = nmlop.OR(criterion, radius).reduce()
    return (retval, [])

varact2vars60x_houses = {
    'old_house_count_town'               : {'var': 0x60, 'start':  0, 'size':  8},
    'old_house_count_map'                : {'var': 0x60, 'start':  8, 'size':  8},
    'other_house_count_town'             : {'var': 0x61, 'start':  0, 'size':  8},
    'other_house_count_map'              : {'var': 0x61, 'start':  8, 'size':  8},
    'other_class_count_town'             : {'var': 0x61, 'start': 16, 'size':  8},
    'other_class_count_map'              : {'var': 0x61, 'start': 24, 'size':  8},
    'nearby_tile_slope'                  : {'var': 0x62, 'start':  0, 'size':  5, 'param_function': signed_tile_offset},
    'nearby_tile_is_water'               : {'var': 0x62, 'start':  9, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_terrain_type'           : {'var': 0x62, 'start': 10, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_water_class'            : {'var': 0x62, 'start': 13, 'size':  2, 'param_function': signed_tile_offset},
    'nearby_tile_height'                 : {'var': 0x62, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_class'                  : {'var': 0x62, 'start': 24, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_animation_frame'        : {'var': 0x63, 'start':  0, 'size':  8, 'param_function': signed_tile_offset},
    'cargo_accepted_nearby_ever'         : {'var': 0x64, 'start':  0, 'size':  1, 'param_function': cargo_accepted_nearby},
    'cargo_accepted_nearby_last_month'   : {'var': 0x64, 'start':  1, 'size':  1, 'param_function': cargo_accepted_nearby},
    'cargo_accepted_nearby_this_month'   : {'var': 0x64, 'start':  2, 'size':  1, 'param_function': cargo_accepted_nearby},
    'cargo_accepted_nearby_last_bigtick' : {'var': 0x64, 'start':  3, 'size':  1, 'param_function': cargo_accepted_nearby},
    'cargo_accepted_nearby_watched'      : {'var': 0x64, 'start':  4, 'size':  1, 'param_function': cargo_accepted_nearby},
    'nearest_house_matching_criterion'   : {'var': 0x65, 'start':  0, 'size':  8, 'param_function': nearest_house_matching_criterion},
    'nearby_tile_house_id'               : {'var': 0x66, 'start':  0, 'size': 16, 'param_function': signed_tile_offset, 'value_function': value_sign_extend},
    'nearby_tile_house_class'            : {'var': 0x66, 'start': 16, 'size': 16, 'param_function': signed_tile_offset, 'value_function': value_sign_extend},
    'nearby_tile_house_grfid'            : {'var': 0x67, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
}

#
# Global variables (feature 0x08) have no variational action2
#

#
# Industry tiles (feature 0x09)
#

varact2vars_industrytiles = {
    'construction_state' : {'var': 0x40, 'start': 0, 'size':  2},
    'terrain_type'       : {'var': 0x41, 'start': 0, 'size':  8},
    'town_zone'          : {'var': 0x42, 'start': 0, 'size':  3},
    'relative_x'         : {'var': 0x43, 'start': 0, 'size':  8},
    'relative_y'         : {'var': 0x43, 'start': 8, 'size':  8},
    'relative_pos'       : {'var': 0x43, 'start': 0, 'size': 16},
    'animation_frame'    : {'var': 0x44, 'start': 0, 'size':  8},
    'random_bits'        : {'var': 0x5F, 'start': 8, 'size':  8},
}

varact2vars60x_industrytiles = {
    'nearby_tile_slope'            : {'var': 0x60, 'start':  0, 'size':  5, 'param_function': signed_tile_offset},
    'nearby_tile_is_same_industry' : {'var': 0x60, 'start':  8, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_is_water'         : {'var': 0x60, 'start':  9, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_terrain_type'     : {'var': 0x60, 'start': 10, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_water_class'      : {'var': 0x60, 'start': 13, 'size':  2, 'param_function': signed_tile_offset},
    'nearby_tile_height'           : {'var': 0x60, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_class'            : {'var': 0x60, 'start': 24, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_animation_frame'  : {'var': 0x61, 'start':  0, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_industrytile_id'  : {'var': 0x62, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
}

#
# Industries (feature 0x0A)
#

varact2vars_industries = {
    'waiting_cargo_1'              : {'var': 0x40, 'start':  0, 'size': 16, 'replaced_by': 'incoming_cargo_waiting'},
    'waiting_cargo_2'              : {'var': 0x41, 'start':  0, 'size': 16, 'replaced_by': 'incoming_cargo_waiting'},
    'waiting_cargo_3'              : {'var': 0x42, 'start':  0, 'size': 16, 'replaced_by': 'incoming_cargo_waiting'},
    'water_distance'               : {'var': 0x43, 'start':  0, 'size': 32},
    'layout_num'                   : {'var': 0x44, 'start':  0, 'size':  8},
    # bits 0 .. 16 are either useless or already covered by var A7
    'founder_type'                 : {'var': 0x45, 'start': 16, 'size':  2},
    'founder_colour1'              : {'var': 0x45, 'start': 24, 'size':  4},
    'founder_colour2'              : {'var': 0x45, 'start': 28, 'size':  4},
    'build_date'                   : {'var': 0x46, 'start':  0, 'size': 32},
    'gs_disallows_prod_decrease'   : {'var': 0x47, 'start':  0, 'size': 1},
    'gs_disallows_prod_increase'   : {'var': 0x47, 'start':  1, 'size': 1},
    'gs_disallows_closure'         : {'var': 0x47, 'start':  2, 'size': 1},
    'random_bits'                  : {'var': 0x5F, 'start':  8, 'size': 16},
    'produced_cargo_waiting_1'     : {'var': 0x8A, 'start':  0, 'size': 16, 'replaced_by': 'produced_cargo_waiting'},
    'produced_cargo_waiting_2'     : {'var': 0x8C, 'start':  0, 'size': 16, 'replaced_by': 'produced_cargo_waiting'},
    'production_rate_1'            : {'var': 0x8E, 'start':  0, 'size':  8, 'replaced_by': 'production_rate'},
    'production_rate_2'            : {'var': 0x8F, 'start':  0, 'size':  8, 'replaced_by': 'production_rate'},
    'production_level'             : {'var': 0x93, 'start':  0, 'size':  8},
    'produced_this_month_1'        : {'var': 0x94, 'start':  0, 'size': 16, 'replaced_by': 'this_month_production'},
    'produced_this_month_2'        : {'var': 0x96, 'start':  0, 'size': 16, 'replaced_by': 'this_month_production'},
    'transported_this_month_1'     : {'var': 0x98, 'start':  0, 'size': 16, 'replaced_by': 'this_month_transported'},
    'transported_this_month_2'     : {'var': 0x9A, 'start':  0, 'size': 16, 'replaced_by': 'this_month_transported'},
    'transported_last_month_pct_1' : {'var': 0x9C, 'start':  0, 'size':  8, 'value_function': value_mul_div(101, 256), 'replaced_by': 'transported_last_month_pct'},
    'transported_last_month_pct_2' : {'var': 0x9D, 'start':  0, 'size':  8, 'value_function': value_mul_div(101, 256), 'replaced_by': 'transported_last_month_pct'},
    'produced_last_month_1'        : {'var': 0x9E, 'start':  0, 'size': 16, 'replaced_by': 'last_month_production'},
    'produced_last_month_2'        : {'var': 0xA0, 'start':  0, 'size': 16, 'replaced_by': 'last_month_production'},
    'transported_last_month_1'     : {'var': 0xA2, 'start':  0, 'size': 16, 'replaced_by': 'last_month_transported'},
    'transported_last_month_2'     : {'var': 0xA4, 'start':  0, 'size': 16, 'replaced_by': 'last_month_transported'},
    'founder'                      : {'var': 0xA7, 'start':  0, 'size':  8},
    'colour'                       : {'var': 0xA8, 'start':  0, 'size':  8},
    'counter'                      : {'var': 0xAA, 'start':  0, 'size': 16},
    'build_type'                   : {'var': 0xB3, 'start':  0, 'size':  2},
    'last_accept_date'             : {'var': 0xB4, 'start':  0, 'size': 16, 'value_function': value_add_constant(701265)},
}

def industry_count(name, args, pos, info):
    if len(args) < 1 or len(args) > 2:
        raise generic.ScriptError("'{}'() requires between 1 and 2 argument(s), encountered {:d}".format(name, len(args)), pos)

    grfid = expression.ConstantNumeric(0xFFFFFFFF) if len(args) == 1 else args[1]
    extra_params = [(0x100, grfid)]

    return (args[0], extra_params)


def industry_layout_count(name, args, pos, info):
    if len(args) < 2 or len(args) > 3:
        raise generic.ScriptError("'{}'() requires between 2 and 3 argument(s), encountered {:d}".format(name, len(args)), pos)

    grfid = expression.ConstantNumeric(0xFFFFFFFF) if len(args) == 2 else args[2]

    extra_params = []
    extra_params.append( (0x100, grfid) )
    extra_params.append( (0x101, nmlop.AND(args[1], 0xFF).reduce()) )
    return (args[0], extra_params)

def industry_town_count(name, args, pos, info):
    if len(args) < 1 or len(args) > 2:
        raise generic.ScriptError("'{}'() requires between 1 and 2 argument(s), encountered {:d}".format(name, len(args)), pos)

    grfid = expression.ConstantNumeric(0xFFFFFFFF) if len(args) == 1 else args[1]

    extra_params = []
    extra_params.append( (0x100, grfid) )
    extra_params.append( (0x101, expression.ConstantNumeric(0x0100)) )
    return (args[0], extra_params)

def industry_cargotype(name, args, pos, info):
    return (expression.functioncall.builtin_resolve_typelabel(name, args, pos, table_name="cargotype"), [])

varact2vars60x_industries = {
    'nearby_tile_industry_tile_id' : {'var': 0x60, 'start':  0, 'size': 16, 'param_function': unsigned_tile_offset},
    'nearby_tile_random_bits'      : {'var': 0x61, 'start':  0, 'size':  8, 'param_function': unsigned_tile_offset},
    'nearby_tile_slope'            : {'var': 0x62, 'start':  0, 'size':  5, 'param_function': unsigned_tile_offset},
    'nearby_tile_is_water'         : {'var': 0x62, 'start':  9, 'size':  1, 'param_function': unsigned_tile_offset},
    'nearby_tile_terrain_type'     : {'var': 0x62, 'start': 10, 'size':  3, 'param_function': unsigned_tile_offset},
    'nearby_tile_water_class'      : {'var': 0x62, 'start': 13, 'size':  2, 'param_function': unsigned_tile_offset},
    'nearby_tile_height'           : {'var': 0x62, 'start': 16, 'size':  8, 'param_function': unsigned_tile_offset},
    'nearby_tile_class'            : {'var': 0x62, 'start': 24, 'size':  4, 'param_function': unsigned_tile_offset},
    'nearby_tile_animation_frame'  : {'var': 0x63, 'start':  0, 'size':  8, 'param_function': unsigned_tile_offset},
    'town_manhattan_dist'          : {'var': 0x65, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
    'town_zone'                    : {'var': 0x65, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'town_euclidean_dist'          : {'var': 0x66, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
    'industry_count'               : {'var': 0x67, 'start': 16, 'size':  8, 'param_function': industry_count},
    'industry_distance'            : {'var': 0x67, 'start':  0, 'size': 16, 'param_function': industry_count},
    'industry_layout_count'        : {'var': 0x68, 'start': 16, 'size':  8, 'param_function': industry_layout_count},
    'industry_layout_distance'     : {'var': 0x68, 'start':  0, 'size': 16, 'param_function': industry_layout_count},
    'industry_town_count'          : {'var': 0x68, 'start': 16, 'size':  8, 'param_function': industry_town_count},
    'produced_cargo_waiting'       : {'var': 0x69, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'this_month_production'        : {'var': 0x6A, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'this_month_transported'       : {'var': 0x6B, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'last_month_production'        : {'var': 0x6C, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'last_month_transported'       : {'var': 0x6D, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'last_cargo_accepted_at'       : {'var': 0x6E, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'incoming_cargo_waiting'       : {'var': 0x6F, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'production_rate'              : {'var': 0x70, 'start':  0, 'size': 32, 'param_function': industry_cargotype},
    'transported_last_month_pct'   : {'var': 0x71, 'start':  0, 'size': 32, 'param_function': industry_cargotype, 'value_function': value_mul_div(101, 256)},
}

#
# Cargos (feature 0x0B) have no own varaction2 variables
# Sounds (feature 0x0C) have no variational action2
#

#
# Airports (feature 0x0D)
#

varact2vars_airports = {
    **varact2vars_base_stations,
    'layout' : {'var': 0x40, 'start': 0, 'size': 32},
}
varact2vars60x_airports = {
    **varact2vars60x_base_stations,
}

#
# New Signals (feature 0x0E) are not implemented in OpenTTD
#

#
# Objects (feature 0x0F)
#

varact2vars_objects = {
    'relative_x'             : {'var': 0x40, 'start':  0, 'size':  8},
    'relative_y'             : {'var': 0x40, 'start':  8, 'size':  8},
    'relative_pos'           : {'var': 0x40, 'start':  0, 'size': 16},

    'terrain_type'           : {'var': 0x41, 'start':  0, 'size':  3},
    'tile_slope'             : {'var': 0x41, 'start':  8, 'size':  5},

    'build_date'             : {'var': 0x42, 'start':  0, 'size': 32},
    'animation_frame'        : {'var': 0x43, 'start':  0, 'size':  8},
    'owner'                  : {'var': 0x44, 'start':  0, 'size':  8},

    'town_manhattan_dist'    : {'var': 0x45, 'start':  0, 'size': 16},
    'town_zone'              : {'var': 0x45, 'start': 16, 'size':  8},

    'town_euclidean_dist'    : {'var': 0x46, 'start':  0, 'size': 16},
    'colour'                 : {'var': 0x47, 'start':  0, 'size':  8},
    'view'                   : {'var': 0x48, 'start':  0, 'size':  8},
    'random_bits'            : {'var': 0x5F, 'start':  8, 'size':  8},
}

varact2vars60x_objects = {
    'nearby_tile_object_type'      : {'var': 0x60, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
    'nearby_tile_object_view'      : {'var': 0x60, 'start': 16, 'size':  4, 'param_function': signed_tile_offset},

    'nearby_tile_random_bits'      : {'var': 0x61, 'start':  0, 'size':  8, 'param_function': signed_tile_offset},

    'nearby_tile_slope'            : {'var': 0x62, 'start':  0, 'size':  5, 'param_function': signed_tile_offset},
    'nearby_tile_is_same_object'   : {'var': 0x62, 'start':  8, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_is_water'         : {'var': 0x62, 'start':  9, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_terrain_type'     : {'var': 0x62, 'start': 10, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_water_class'      : {'var': 0x62, 'start': 13, 'size':  2, 'param_function': signed_tile_offset},
    'nearby_tile_height'           : {'var': 0x62, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_class'            : {'var': 0x62, 'start': 24, 'size':  4, 'param_function': signed_tile_offset},

    'nearby_tile_animation_frame'  : {'var': 0x63, 'start':  0, 'size':  8, 'param_function': signed_tile_offset},

    'object_count'                 : {'var': 0x64, 'start': 16, 'size':  8, 'param_function': industry_count},
    'object_distance'              : {'var': 0x64, 'start':  0, 'size': 16, 'param_function': industry_count},
}

#
# Railtypes (feature 0x10)
#

varact2vars_railtype = {
    'terrain_type'          : {'var': 0x40, 'start': 0, 'size':  8},
    'enhanced_tunnels'      : {'var': 0x41, 'start': 0, 'size':  8},
    'level_crossing_status' : {'var': 0x42, 'start': 0, 'size':  8},
    'build_date'            : {'var': 0x43, 'start': 0, 'size': 32},
    'town_zone'             : {'var': 0x44, 'start': 0, 'size':  8},
    'roadtype'              : {'var': 0x45, 'start': 0, 'size':  8},
    'tramtype'              : {'var': 0x45, 'start': 8, 'size':  8},
    'railtype'              : {'var': 0x45, 'start': 16, 'size': 8},
    'random_bits'           : {'var': 0x5F, 'start': 8, 'size':  2},
}
# Railtypes have no 60+x variables

#
# Airport tiles (feature 0x11)
#

varact2vars_airporttiles = {
    'terrain_type'        : {'var': 0x41, 'start':  0, 'size':  8},
    'town_radius_group'   : {'var': 0x42, 'start':  0, 'size':  3},
    'relative_x'          : {'var': 0x43, 'start':  0, 'size':  8},
    'relative_y'          : {'var': 0x43, 'start':  8, 'size':  8},
    'relative_pos'        : {'var': 0x43, 'start':  0, 'size': 16},
    'animation_frame'     : {'var': 0x44, 'start':  0, 'size':  8},
    'random_bits_station' : {'var': 0x5F, 'start':  8, 'size': 16},
    'random_bits_tile'    : {'var': 0x5F, 'start': 24, 'size':  4},
}

varact2vars60x_airporttiles = {
    'nearby_tile_slope'            : {'var': 0x60, 'start':  0, 'size':  5, 'param_function': signed_tile_offset},
    'nearby_tile_is_same_airport'  : {'var': 0x60, 'start':  8, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_is_water'         : {'var': 0x60, 'start':  9, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_terrain_type'     : {'var': 0x60, 'start': 10, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_water_class'      : {'var': 0x60, 'start': 13, 'size':  2, 'param_function': signed_tile_offset},
    'nearby_tile_height'           : {'var': 0x60, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_class'            : {'var': 0x60, 'start': 24, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_animation_frame'  : {'var': 0x61, 'start':  0, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_airporttile_id'   : {'var': 0x62, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
}

#
# Roadtypes (feature 0x12)
#

varact2vars_roadtype = {
    'terrain_type'          : {'var': 0x40, 'start': 0, 'size':  8},
    'enhanced_tunnels'      : {'var': 0x41, 'start': 0, 'size':  8},
    'level_crossing_status' : {'var': 0x42, 'start': 0, 'size':  8},
    'build_date'            : {'var': 0x43, 'start': 0, 'size': 32},
    'town_zone'             : {'var': 0x44, 'start': 0, 'size':  8},
    'roadtype'              : {'var': 0x45, 'start': 0, 'size':  8},
    'tramtype'              : {'var': 0x45, 'start': 8, 'size':  8},
    'railtype'              : {'var': 0x45, 'start': 16, 'size': 8},
    'random_bits'           : {'var': 0x5F, 'start': 8, 'size':  2},
}
# Roadtypes have no 60+x variables

#
# Tramtypes (feature 0x13)
#

varact2vars_tramtype = {
    'terrain_type'          : {'var': 0x40, 'start': 0, 'size':  8},
    'enhanced_tunnels'      : {'var': 0x41, 'start': 0, 'size':  8},
    'level_crossing_status' : {'var': 0x42, 'start': 0, 'size':  8},
    'build_date'            : {'var': 0x43, 'start': 0, 'size': 32},
    'town_zone'             : {'var': 0x44, 'start': 0, 'size':  8},
    'roadtype'              : {'var': 0x45, 'start': 0, 'size':  8},
    'tramtype'              : {'var': 0x45, 'start': 8, 'size':  8},
    'railtype'              : {'var': 0x45, 'start': 16, 'size': 8},
    'random_bits'           : {'var': 0x5F, 'start': 8, 'size':  2},
}
# Tramtypes have no 60+x variables


#
# Towns are not a true feature, but accessible via the parent scope of e.g. industries, stations
#

varact2vars_towns = {
    'is_city'                        : {'var': 0x40, 'start': 0, 'size': 1},
    'cities_enabled'                 : {'var': 0x40, 'start': 1, 'size': 1, 'value_function': lambda var, info: expression.Not(var, var.pos)},
    'town_index'                     : {'var': 0x41, 'start': 0, 'size': 16},
    'population'                     : {'var': 0x82, 'start': 0, 'size': 16},
    'has_church'                     : {'var': 0x92, 'start': 1, 'size': 1},
    'has_stadium'                    : {'var': 0x92, 'start': 2, 'size': 1},
    'town_zone_0_radius_square'      : {'var': 0x94, 'start': 0, 'size': 16},
    'town_zone_1_radius_square'      : {'var': 0x96, 'start': 0, 'size': 16},
    'town_zone_2_radius_square'      : {'var': 0x98, 'start': 0, 'size': 16},
    'town_zone_3_radius_square'      : {'var': 0x9A, 'start': 0, 'size': 16},
    'town_zone_4_radius_square'      : {'var': 0x9C, 'start': 0, 'size': 16},
    'num_houses'                     : {'var': 0xB6, 'start': 0, 'size': 16},
    'percent_transported_passengers' : {'var': 0xCA, 'start': 0, 'size': 8, 'value_function': value_mul_div(101, 256)},
    'percent_transported_mail'       : {'var': 0xCB, 'start': 0, 'size': 8, 'value_function': value_mul_div(101, 256)},
}

#
# Roadstops (feature road_stops)
#

varact2vars_roadstop = {
    **varact2vars_base_stations,
    'view'                  : {'var': 0x40, 'start':  0, 'size':  8},

    'stop_type'             : {'var': 0x41, 'start':  0, 'size':  8},

    'terrain_type'          : {'var': 0x42, 'start':  0, 'size':  8},
    'tile_slope'            : {'var': 0x42, 'start':  8, 'size':  5},

    'has_road'              : {'var': 0x43, 'start':  0, 'size': 32, 'value_function': lambda var, info: nmlop.CMP_NEQ(var, 0xFFFFFFFF)},
    'has_tram'              : {'var': 0x44, 'start':  0, 'size': 32, 'value_function': lambda var, info: nmlop.CMP_NEQ(var, 0xFFFFFFFF)},
    'road_type'             : {'var': 0x43, 'start':  0, 'size':  8}, # The roadtype of this tile
    'tram_type'             : {'var': 0x44, 'start':  0, 'size':  8}, # The tramtype of this tile

    'town_manhattan_dist'   : {'var': 0x45, 'start':  0, 'size': 16},
    'town_zone'             : {'var': 0x45, 'start': 16, 'size':  8},
    'town_euclidean_dist'   : {'var': 0x46, 'start':  0, 'size': 32},

    'company_num'           : {'var': 0x47, 'start':  0, 'size':  8}, # 0..14 company number
    'company_type'          : {'var': 0x47, 'start': 16, 'size':  2}, # PLAYERTYPE_HUMAN, PLAYERTYPE_AI etc.
    'company_colour1'       : {'var': 0x47, 'start': 24, 'size':  4}, # COLOUR_XXX. See https://newgrf-specs.tt-wiki.net/wiki/NML:List_of_default_colour_translation_palettes#Company_colour_helper_functions
    'company_colour2'       : {'var': 0x47, 'start': 28, 'size':  4}, # Same as above

    'animation_frame'       : {'var': 0x49, 'start':  0, 'size':  8},

    'drawn_in_gui'          : {'var': 0x50, 'start':  4, 'size':  1},

    'waiting_triggers'      : {'var': 0x5F, 'start':  0, 'size':  8},
    'random_bits_tile'      : {'var': 0x5F, 'start': 24, 'size':  8},
}

varact2vars60x_roadstop = {
    **varact2vars60x_base_stations,
    'nearby_tile_animation_frame'       : {'var': 0x66, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
    'nearby_tile_info'                  : {'var': 0x67, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
    'nearby_tile_slope'                 : {'var': 0x67, 'start':  0, 'size':  5, 'param_function': signed_tile_offset},
    'nearby_tile_is_water'              : {'var': 0x67, 'start':  9, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_terrain_type'          : {'var': 0x67, 'start': 10, 'size':  3, 'param_function': signed_tile_offset},
    'nearby_tile_water_class'           : {'var': 0x67, 'start': 13, 'size':  2, 'param_function': signed_tile_offset},
    'nearby_tile_height'                : {'var': 0x67, 'start': 16, 'size':  8, 'param_function': signed_tile_offset},
    'nearby_tile_class'                 : {'var': 0x67, 'start': 24, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_road_stop_info'        : {'var': 0x68, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
    'nearby_tile_is_road_stop'          : {'var': 0x68, 'start':  0, 'size': 32, 'param_function': signed_tile_offset, 'value_function': lambda var, info: nmlop.CMP_NEQ(var, 0xFFFFFFFF)},
    'nearby_tile_same_grf'              : {'var': 0x68, 'start':  8, 'size':  2, 'param_function': signed_tile_offset, 'value_function': value_equals(0)},
    'nearby_tile_other_grf'             : {'var': 0x68, 'start':  8, 'size':  2, 'param_function': signed_tile_offset, 'value_function': value_equals(1)},
    'nearby_tile_original_gfx'          : {'var': 0x68, 'start':  8, 'size':  2, 'param_function': signed_tile_offset, 'value_function': value_equals(2)},
    'nearby_tile_same_station'          : {'var': 0x68, 'start': 10, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_different_view'        : {'var': 0x68, 'start': 11, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_view'                  : {'var': 0x68, 'start': 12, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_is_drive_through'      : {'var': 0x68, 'start': 12, 'size':  4, 'param_function': signed_tile_offset, 'value_function': lambda var, info: nmlop.CMP_GE(var, 4)},
    'nearby_tile_stop_type'             : {'var': 0x68, 'start': 16, 'size':  4, 'param_function': signed_tile_offset},
    'nearby_tile_same_stop_type'        : {'var': 0x68, 'start': 20, 'size':  1, 'param_function': signed_tile_offset},
    'nearby_tile_grfid'                 : {'var': 0x6A, 'start':  0, 'size': 32, 'param_function': signed_tile_offset},
    'nearby_tile_road_stop_id'          : {'var': 0x6B, 'start':  0, 'size': 16, 'param_function': signed_tile_offset},
}

class VarAct2Scope:
    def __init__(self, name, vars_normal, vars_60x, has_persistent_storage=False):
        self.name = name
        self.vars_normal = vars_normal
        self.vars_60x = vars_60x
        self.has_persistent_storage = has_persistent_storage


class VarAct2Feature:
    def __init__(self, self_scope, parent_scope):
        self.self_scope = self_scope
        self.parent_scope = parent_scope

    def get_scope(self, var_range):
        assert var_range in (0x89, 0x8A)
        return self.self_scope if var_range == 0x89 else self.parent_scope


scope_towns = VarAct2Scope("Towns", varact2vars_towns, {}, has_persistent_storage=True)
scope_trains = VarAct2Scope("Trains", varact2vars_trains, varact2vars60x_trains)
scope_roadvehs = VarAct2Scope("RoadVehs", varact2vars_roadvehs, varact2vars60x_roadvehs)
scope_ships = VarAct2Scope("Ships", varact2vars_ships, varact2vars60x_vehicles)
scope_aircraft = VarAct2Scope("Aircraft", varact2vars_aircraft, varact2vars60x_vehicles)
scope_stations = VarAct2Scope("Stations", varact2vars_stations, varact2vars60x_stations)
scope_canals = VarAct2Scope("Canals", varact2vars_canals, {})
scope_houses = VarAct2Scope("Houses", varact2vars_houses, varact2vars60x_houses)
scope_industrytiles = VarAct2Scope("IndustryTiles", varact2vars_industrytiles, varact2vars60x_industrytiles)
scope_industries = VarAct2Scope(
    "Industries", varact2vars_industries, varact2vars60x_industries, has_persistent_storage=True
)
scope_cargos = VarAct2Scope("Cargos", {}, {})
scope_soundeffects = VarAct2Scope("SoundEffects", {}, {})
scope_airports = VarAct2Scope("Airports", varact2vars_airports, varact2vars60x_airports, has_persistent_storage=True)
scope_objects = VarAct2Scope("Objects", varact2vars_objects, varact2vars60x_objects)
scope_railtypes = VarAct2Scope("RailTypes", varact2vars_railtype, {})
scope_airporttiles = VarAct2Scope("AirportTiles", varact2vars_airporttiles, varact2vars60x_airporttiles)
scope_roadtypes = VarAct2Scope("RoadTypes", varact2vars_roadtype, {})
scope_tramtypes = VarAct2Scope("TramTypes", varact2vars_tramtype, {})
scope_roadstops = VarAct2Scope("RoadStops", varact2vars_roadstop, varact2vars60x_roadstop)

varact2features = [
    VarAct2Feature(scope_trains, scope_trains),
    VarAct2Feature(scope_roadvehs, scope_roadvehs),
    VarAct2Feature(scope_ships, scope_ships),
    VarAct2Feature(scope_aircraft, scope_aircraft),
    VarAct2Feature(scope_stations, scope_towns),
    VarAct2Feature(scope_canals, None),
    None,  # bridges
    VarAct2Feature(scope_houses, scope_towns),
    None,  # globalvars
    VarAct2Feature(scope_industrytiles, scope_industries),
    VarAct2Feature(scope_industries, scope_towns),
    VarAct2Feature(scope_cargos, None),
    VarAct2Feature(scope_soundeffects, None),
    VarAct2Feature(scope_airports, scope_towns),
    None,  # signals
    VarAct2Feature(scope_objects, scope_towns),
    VarAct2Feature(scope_railtypes, None),
    VarAct2Feature(scope_airporttiles, scope_airports),
    VarAct2Feature(scope_roadtypes, None),
    VarAct2Feature(scope_tramtypes, None),
    VarAct2Feature(scope_roadstops, scope_towns),
]



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import array
import json
import os

from nml import generic

keep_orphaned = True


class SpriteCache:
    """
    Cache for compressed sprites.

    @ivar sources: Tuple of paths to files the cache belongs to or depends on.
    @type sources: C{tuple} of (C{str} or C{None})

    @ivar cache_time: Date of cache files. The cache is invalid, if the source image files are newer.
    @type cache_time: C{int}

    @ivar cached_sprites: Cache contents
    @type cached_sprites: C{dict} mapping cache keys to cache items.

    Cache file format description:
        Format of cache index is JSON (JavaScript Object Notation), which is
        easily readable by both humans (for debugging) and machines. Format is as follows:

        A list of dictionaries, each corresponding to a sprite, with the following keys
         - rgb_file: filename of the 32bpp sprite (string)
         - rgb_rect: (uncropped) rectangle of the 32bpp sprite (list with 4 elements (x,y,w,h))
         - mask_file, mask_rect: same as above, but for 8bpp sprite
         - mask_pal: palette of mask file, 'DEFAULT' or 'LEGACY'. Not present if 'mask_file' is not present.
         - crop: List of 4 positive integers, indicating how much to crop if cropping is enabled
              Order is (left, right, top, bottom), it is not present if cropping is disabled
         - info: Info byte of the sprite
         - pixel_stats: Dictionary with statistics about pixels:
              'total': Total amount of pixels.
              'alpha': Amount of semi-transparent pixels in 32bpp.
              'white': Amount of pure-white pixels in 8bpp.
              'anim':  Amount of animated pixels in 8bpp.
         - offset: Offset into the cache file for this sprite
         - size: Length of this sprite in the cache file

        Either rgb_file/rect, mask_file/rect, or both must be specified, depending on the sprite
        The cache should contain the sprite data, but not the header (sizes/offsets and such)

        For easy lookup, this information is stored in a dictionary
        Tuples are used because these are hashable

        The key is a (rgb_file, rgb_rect, mask_file, mask_rect, do_crop, palette)-tuple
        The rectangles are 4-tuples, non-existent items (e.g. no rgb) are None
        do_crop is a boolean indicating if this sprite has been cropped
        palette is a string identifier

        The value that this key maps to is a 6-tuple, containing:
         - the sprite data (as a byte array)
         - The 'info' byte of the sprite
         - The cropping information (see above) (None if 'do_crop' in the key is false)
         - The pixel_stats dictionary with statistics.
         - Whether the sprite exists in the old (loaded)cache
         - whether the sprite is used by the current GRF

        The cache file itself is simply the binary sprite data with no
        meta-information or padding. Offsets and sizes for the various sprites
        are in the cacheindex file.
    """

    def __init__(self, sources=()):
        self.sources = sources
        self.cache_time = 0
        self.cached_sprites = {}

    def get_item(self, cache_key, palette):
        """
        Get item from cache.

        @param cache_key: Sprite metadata key, as returned by RealSprite.get_cache_key
        @type  cache_key: C{tuple}

        @param palette: Palette identifier, one of palette.palette_name
        @type  palette: C{str}

        @return: Cache item
        @rtype: C{tuple} or C{None}
        """

        if cache_key[2] is not None:
            key = cache_key + (palette,)
        else:
            key = cache_key + (None,)
        return self.cached_sprites.get(key, None)

    def add_item(self, cache_key, palette, item):
        """
        Add item to cache.

        @param cache_key: Sprite metadata key, as returned by RealSprite.get_cache_key
        @type  cache_key: C{tuple}

        @param palette: Palette identifier, one of palette.palette_name
        @type  palette: C{str}

        @param item: Cache item
        @type  item: C{tuple}
        """

        if cache_key[2] is not None:
            key = cache_key + (palette,)
        else:
            key = cache_key + (None,)
        self.cached_sprites[key] = item

    def count_orphaned(self):
        """
        Count number of items in cache, which have not been used.

        @return: Number of unused items.
        @type:   C{int}
        """

        return sum(not item[5] for item in self.cached_sprites.values())

    def read_cache(self):
        """
        Read the *.grf.cache[index] files.
        """

        try:
            with generic.open_cache_file(self.sources, ".cache", "rb") as cache_file:
                cache_data = array.array("B")
                cache_size = os.fstat(cache_file.fileno()).st_size
                cache_data.fromfile(cache_file, cache_size)
                assert cache_size == len(cache_data)
                self.cache_time = os.path.getmtime(cache_file.name)

            with generic.open_cache_file(self.sources, ".cacheindex", "r") as index_file:
                index_file_name = index_file.name
                sprite_index = json.load(index_file)
        except OSError:
            # Cache files don't exist (or otherwise aren't readable)
            return
        except json.JSONDecodeError:
            generic.print_warning(
                generic.Warning.GENERIC,
                "{} contains invalid data, ignoring.".format(index_file_name)
                + " Please remove the file and file a bug report if this warning keeps appearing",
            )
            self.cached_sprites = {}
            return

        source_mtime = {}

        try:
            # Just assert and print a generic message on errors, as the cache data should be correct
            # Not asserting could lead to errors later on
            # Also, it doesn't make sense to inform the user about things he shouldn't know about and can't fix
            assert isinstance(sprite_index, list)
            for sprite in sprite_index:
                assert isinstance(sprite, dict)
                # load RGB (32bpp) data
                rgb_key = (None, None)
                if "rgb_file" in sprite and "rgb_rect" in sprite:
                    assert isinstance(sprite["rgb_file"], str)
                    assert isinstance(sprite["rgb_rect"], list) and len(sprite["rgb_rect"]) == 4
                    assert all(isinstance(num, int) for num in sprite["rgb_rect"])
                    rgb_key = (sprite["rgb_file"], tuple(sprite["rgb_rect"]))

                # load Mask (8bpp) data
                mask_key = (None, None)
                if "mask_file" in sprite and "mask_rect" in sprite:
                    assert isinstance(sprite["mask_file"], str)
                    assert isinstance(sprite["mask_rect"], list) and len(sprite["mask_rect"]) == 4
                    assert all(isinstance(num, int) for num in sprite["mask_rect"])
                    mask_key = (sprite["mask_file"], tuple(sprite["mask_rect"]))

                palette_key = None
                if "mask_pal" in sprite:
                    palette_key = sprite["mask_pal"]

                # Compose key
                assert any(i is not None for i in rgb_key + mask_key)
                key = rgb_key + mask_key + ("crop" in sprite, palette_key)
                assert key not in self.cached_sprites

                # Read size/offset from cache
                assert "offset" in sprite and "size" in sprite
                offset, size = sprite["offset"], sprite["size"]
                assert isinstance(offset, int) and isinstance(size, int)
                assert offset >= 0 and size > 0
                assert offset + size <= cache_size
                data = cache_data[offset : offset + size]

                # Read info / cropping data from cache
                assert "info" in sprite and isinstance(sprite["info"], int)
                info = sprite["info"]
                if "crop" in sprite:
                    assert isinstance(sprite["crop"], list) and len(sprite["crop"]) == 4
                    assert all(isinstance(num, int) for num in sprite["crop"])
                    crop = tuple(sprite["crop"])
                else:
                    crop = None

                if "pixel_stats" in sprite:
                    assert isinstance(sprite["pixel_stats"], dict)
                    pixel_stats = sprite["pixel_stats"]
                else:
                    pixel_stats = {}

                # Compose value
                value = (data, info, crop, pixel_stats, True, False)

                # Check if cache item is still valid
                is_valid = True
                if rgb_key[0] is not None:
                    mtime = source_mtime.get(rgb_key[0])
                    if mtime is None:
                        mtime = os.path.getmtime(generic.find_file(rgb_key[0]))
                        source_mtime[rgb_key[0]] = mtime

                    if mtime > self.cache_time:
                        is_valid = False

                if mask_key[0] is not None:
                    mtime = source_mtime.get(mask_key[0])
                    if mtime is None:
                        mtime = os.path.getmtime(generic.find_file(mask_key[0]))
                        source_mtime[mask_key[0]] = mtime

                    if mtime > self.cache_time:
                        is_valid = False

                # Drop items from older spritecache format without palette entry
                if (mask_key[0] is None) != (palette_key is None):
                    is_valid = False

                if is_valid:
                    self.cached_sprites[key] = value
        except Exception:
            generic.print_warning(
                generic.Warning.GENERIC,
                "{} contains invalid data, ignoring.".format(index_file_name)
                + " Please remove the file and file a bug report if this warning keeps appearing",
            )
            self.cached_sprites = {}  # Clear cache

    def write_cache(self):
        """
        Write the cache data to the .cache[index] files.
        """
        if generic.cache_root_dir is None:
            # Writing cache files will fail, so bail early.
            return

        index_data = []
        sprite_data = array.array("B")
        offset = 0

        old_cache_valid = True
        for key, value in self.cached_sprites.items():
            # Unpack key/value
            rgb_file, rgb_rect, mask_file, mask_rect, do_crop, mask_pal = key
            data, info, crop_rect, pixel_stats, in_old_cache, in_use = value
            assert do_crop == (crop_rect is not None)
            assert (mask_file is None) == (mask_pal is None)

            # If this cache information is exactly the same as the old cache, then we don't bother writing later on
            if not in_use and not keep_orphaned:
                old_cache_valid = False
                continue
            if not in_old_cache:
                old_cache_valid = False

            # Create dictionary with information
            sprite = {}
            if rgb_file is not None:
                sprite["rgb_file"] = rgb_file
                sprite["rgb_rect"] = tuple(rgb_rect)
            if mask_file is not None:
                sprite["mask_file"] = mask_file
                sprite["mask_rect"] = tuple(mask_rect)
                sprite["mask_pal"] = mask_pal

            size = len(data)
            sprite["offset"] = offset
            sprite["size"] = size
            sprite["info"] = info
            if do_crop:
                sprite["crop"] = tuple(crop_rect)
            sprite["pixel_stats"] = pixel_stats

            index_data.append(sprite)
            sprite_data.extend(data)
            offset += size

        if old_cache_valid:
            return

        index_output = json.JSONEncoder(sort_keys=True).encode(index_data)

        try:
            with generic.open_cache_file(self.sources, ".cache", "wb") as cache_file, generic.open_cache_file(
                self.sources, ".cacheindex", "w"
            ) as index_file:
                index_file.write(index_output)
                sprite_data.tofile(cache_file)
        except OSError:
            return



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action0
from nml.ast import assignment, base_statement


class BaseTracktypeTable(base_statement.BaseStatement):
    """Base class for RailtypeTable, RoadtypeTable etc."""

    def __init__(self, tracktype_list, pos):
        base_statement.BaseStatement.__init__(self, self.track_kind + "type table", pos, False, False)
        generic.OnlyOnce.enforce(self, self.track_kind + "type table")
        self.tracktype_table.clear()
        self.tracktype_list = tracktype_list

    def register_names(self):
        for i, tracktype in enumerate(self.tracktype_list):
            if isinstance(tracktype, assignment.Assignment):
                name = tracktype.name
                val_list = []
                for rt in tracktype.value:
                    if isinstance(rt, expression.Identifier):
                        val_list.append(expression.StringLiteral(rt.value, rt.pos))
                    else:
                        val_list.append(rt)
                    expression.parse_string_to_dword(
                        val_list[-1]
                    )  # we don't care about the result, only validate the input
                self.tracktype_list[i] = val_list if len(val_list) > 1 else val_list[0]
            else:
                name = tracktype
                if isinstance(tracktype, expression.Identifier):
                    self.tracktype_list[i] = expression.StringLiteral(tracktype.value, tracktype.pos)
                expression.parse_string_to_dword(
                    self.tracktype_list[i]
                )  # we don't care about the result, only validate the input
            self.tracktype_table[name.value] = i

    def pre_process(self):
        pass

    def debug_print(self, indentation):
        generic.print_dbg(indentation, self.track_kind.title() + "type table")
        for tracktype in self.tracktype_list:
            if isinstance(tracktype, assignment.Assignment):
                generic.print_dbg(indentation + 2, self.track_kind.title() + ":", tracktype.name.value)
                for v in tracktype.value:
                    generic.print_dbg(indentation + 4, "Try:", v.value)
            else:
                generic.print_dbg(indentation + 2, self.track_kind.title() + ":", tracktype.value)

    def get_action_list(self):
        return action0.get_tracktypelist_action(
            self.table_prop_id, self.cond_tracktype_not_defined, self.tracktype_list
        )

    def __str__(self):
        lines = []
        for tracktype in self.tracktype_list:
            if isinstance(tracktype, assignment.Assignment):
                ids = [expression.identifier_to_print(v.value) for v in tracktype.value]
                lines.append("{}: [{}]".format(str(tracktype.name), ", ".join(ids)))
            else:
                lines.append(expression.identifier_to_print(tracktype.value))

        ret = self.track_kind + "typetable {\n    "
        ret += ", ".join(lines)
        ret += "\n}\n"
        return ret


class RailtypeTable(BaseTracktypeTable):
    track_kind = "rail"
    tracktype_table = global_constants.railtype_table
    table_prop_id = 0x12
    cond_tracktype_not_defined = 0x0D

    def __init__(self, *args, **kwargs):
        global_constants.is_default_railtype_table = False
        super().__init__(*args, **kwargs)


class RoadtypeTable(BaseTracktypeTable):
    track_kind = "road"
    tracktype_table = global_constants.roadtype_table
    table_prop_id = 0x16
    cond_tracktype_not_defined = 0x0F

    def __init__(self, *args, **kwargs):
        global_constants.is_default_roadtype_table = False
        super().__init__(*args, **kwargs)


class TramtypeTable(BaseTracktypeTable):
    track_kind = "tram"
    tracktype_table = global_constants.tramtype_table
    table_prop_id = 0x17
    cond_tracktype_not_defined = 0x11

    def __init__(self, *args, **kwargs):
        global_constants.is_default_tramtype_table = False
        super().__init__(*args, **kwargs)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, global_constants
from nml.actions import action0
from nml.ast import base_statement, general


class DisableItem(base_statement.BaseStatement):
    """
    Class representing a 'disable_item' statement in the AST.

    @ivar feature: Feature of the items to disable
    @type feature: L{ConstantNumeric}

    @ivar first_id: First item ID to disable
    @type first_id: L{ConstantNumeric}, or C{None} if not set

    @ivar last_id: Last item ID to disable
    @type last_id: L{ConstantNumeric}, or C{None} if not set
    """

    def __init__(self, param_list, pos):
        base_statement.BaseStatement.__init__(self, "disable_item()", pos)
        if not (1 <= len(param_list) <= 3):
            raise generic.ScriptError(
                "disable_item() requires between 1 and 3 parameters, encountered {:d}.".format(len(param_list)), pos
            )
        self.feature = general.parse_feature(param_list[0])
        self.first_id = param_list[1] if len(param_list) > 1 else None
        self.last_id = param_list[2] if len(param_list) > 2 else None

    def pre_process(self):
        if self.first_id is not None:
            self.first_id = self.first_id.reduce_constant(global_constants.const_list)

        if self.last_id is not None:
            self.last_id = self.last_id.reduce_constant(global_constants.const_list)
            if self.last_id.value < self.first_id.value:
                raise generic.ScriptError("Last id to disable may not be lower than the first id.", self.last_id.pos)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Disable items, feature=" + str(self.feature.value))
        if self.first_id is not None:
            generic.print_dbg(indentation + 2, "First ID:")
            self.first_id.debug_print(indentation + 4)
        if self.last_id is not None:
            generic.print_dbg(indentation + 2, "Last ID:")
            self.last_id.debug_print(indentation + 4)

    def __str__(self):
        ret = str(self.feature)
        if self.first_id is not None:
            ret += ", " + str(self.first_id)
        if self.last_id is not None:
            ret += ", " + str(self.last_id)
        return "disable_item({});\n".format(ret)

    def get_action_list(self):
        return action0.get_disable_actions(self)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

# -*- coding: utf-8 -*-
import io

from nml import generic, grfstrings, output_base
from nml.actions import real_sprite

zoom_levels = {
    0: "normal",
    1: "zi4",
    2: "zi2",
    3: "zo2",
    4: "zo4",
    5: "zo8",
}

bit_depths = {
    8: "8bpp",
    32: "32bpp",
}


class OutputNFO(output_base.SpriteOutputBase):
    def __init__(self, filename, start_sprite_num):
        output_base.SpriteOutputBase.__init__(self, filename)
        self.sprite_num = start_sprite_num

    def open(self):
        self.file = io.StringIO()

    def open_file(self):
        handle = open(self.filename, "w", encoding="utf-8")
        handle.write(
            "// Automatically generated by GRFCODEC. Do not modify!\n"
            "// (Info version 32)\n"
            "// Escapes: 2+ 2- 2< 2> 2u< 2u> 2/ 2% 2u/ 2u% 2* 2& 2| 2^"
            " 2sto = 2s 2rst = 2r 2psto 2ror = 2rot 2cmp 2ucmp 2<< 2u>> 2>>\n"
            "// Escapes: 71 70 7= 7! 7< 7> 7G 7g 7gG 7GG 7gg 7c 7C\n"
            "// Escapes: D= = DR D+ = DF D- = DC Du* = DM D* = DnF Du<< = DnC D<< = DO D& D| Du/ D/ Du% D%\n"
            "// Format: spritenum imagefile depth xpos ypos xsize ysize xrel yrel zoom flags\n\n"
        )
        return handle

    def assemble_file(self, real_file):
        # All print functions add a space in case there's something written after so remove trailing whitespaces
        real_file.write(self.file.getvalue().replace(" \n", "\n"))

    def print_byte(self, value):
        value = self.prepare_byte(value)
        self.file.write("\\b" + str(value) + " ")

    def print_bytex(self, value, pretty_print=None):
        value = self.prepare_byte(value)
        if pretty_print is not None:
            self.file.write(pretty_print + " ")
            return
        self.file.write("{:02X} ".format(value))

    def print_word(self, value):
        value = self.prepare_word(value)
        self.file.write("\\w{:d} ".format(value))

    def print_wordx(self, value):
        value = self.prepare_word(value)
        self.file.write("\\wx{:04X} ".format(value))

    def print_dword(self, value):
        value = self.prepare_dword(value)
        self.file.write("\\d{:d} ".format(value))

    def print_dwordx(self, value):
        value = self.prepare_dword(value)
        self.file.write("\\dx{:08X} ".format(value))

    def print_string(self, value, final_zero=True, force_ascii=False):
        assert self.in_sprite
        self.file.write('"')
        if not grfstrings.is_ascii_string(value):
            if force_ascii:
                raise generic.ScriptError("Expected ascii string but got a unicode string")
            self.file.write("Þ")  # b'\xC3\x9E'.decode('utf-8')
        self.file.write(value.replace('"', '\\"'))
        self.byte_count += grfstrings.get_string_size(value, final_zero, force_ascii)
        self.file.write('" ')
        if final_zero:
            self.print_bytex(0)
            # get_string_size already includes the final 0 byte
            # but print_bytex also increases byte_count, so decrease
            # it here by one to correct it.
            self.byte_count -= 1

    def print_decimal(self, value):
        assert self.in_sprite
        self.file.write(str(value) + " ")

    def newline(self, msg="", prefix="\t"):
        if msg != "":
            msg = prefix + "// " + msg
        self.file.write(msg + "\n")

    def comment(self, msg):
        self.file.write("// " + msg + "\n")

    def start_sprite(self, size, is_real_sprite=False):
        output_base.SpriteOutputBase.start_sprite(self, size, is_real_sprite)
        self.print_decimal(self.sprite_num)
        self.sprite_num += 1
        if not is_real_sprite:
            self.file.write("* ")
            self.print_decimal(size)

    def print_sprite(self, sprite_list):
        """
        @param sprite_list: List of non-empty real sprites for various bit depths / zoom levels
        @type  sprite_list: C{list} of L{RealSprite}
        """
        self.start_sprite(0, True)
        for i, sprite_info in enumerate(sprite_list):
            self.file.write(sprite_info.file.value + " ")
            self.file.write(bit_depths[sprite_info.bit_depth] + " ")
            self.print_decimal(sprite_info.xpos.value)
            self.print_decimal(sprite_info.ypos.value)
            self.print_decimal(sprite_info.xsize.value)
            self.print_decimal(sprite_info.ysize.value)
            self.print_decimal(sprite_info.xrel.value)
            self.print_decimal(sprite_info.yrel.value)
            self.file.write(zoom_levels[sprite_info.zoom_level] + " ")
            if (sprite_info.flags.value & real_sprite.FLAG_NOCROP) != 0:
                self.file.write("nocrop ")
            if sprite_info.mask_file is not None:
                self.newline()
                self.file.write("|\t")
                self.file.write(sprite_info.mask_file.value)
                self.file.write(" mask ")
                mask_x, mask_y = (
                    sprite_info.mask_pos if sprite_info.mask_pos is not None else (sprite_info.xpos, sprite_info.ypos)
                )
                self.print_decimal(mask_x.value)
                self.print_decimal(mask_y.value)
            if i + 1 < len(sprite_list):
                self.newline()
                self.file.write("|\t")
        self.end_sprite()

    def print_empty_realsprite(self):
        self.start_sprite(1)
        self.print_bytex(0)
        self.end_sprite()

    def print_named_filedata(self, filename):
        self.start_sprite(0, True)
        self.file.write("** " + filename)
        self.end_sprite()



#!/usr/bin/env python3

from setuptools import Distribution, Extension, find_packages, setup

import contextlib
import os

try:
    # Update the version by querying git if possible.
    from nml import version_update

    NML_VERSION = version_update.get_and_write_version()
except ImportError:
    # version_update is excluded from released tarballs, so that
    #  only the predetermined version is used when building from one.
    from nml import version_info

    NML_VERSION = version_info.get_nml_version()

default_dist = Distribution()
default_build_py = default_dist.get_command_class('build_py')
default_clean = default_dist.get_command_class('clean')

setup(
    name="nml",
    version=NML_VERSION,
    packages=find_packages(),
    description="An OpenTTD NewGRF compiler for the nml language",
    long_description=(
        "A tool to compile NewGRFs for OpenTTD from nml files"
        "NML is a meta-language that aims to be a lot simpler to"
        " learn and use than nfo used traditionally to write NewGRFs."
    ),
    license="GPL-2.0+",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Compilers",
    ],
    url="https://github.com/OpenTTD/nml",
    author="NML Development Team",
    author_email="nml-team@openttdcoop.org",
    entry_points={"console_scripts": ["nmlc = nml.main:run"]},
    ext_modules=[Extension("nml_lz77", ["nml/_lz77.c"], optional=True)],
    python_requires=">=3.5",
    install_requires=[
        "Pillow>=3.4",
    ],
)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import ConstantNumeric, Expression, Type
from .parameter import parse_string_to_dword

storage_op_info = {
    "STORE_PERM": {"store": True, "perm": True, "grfid": False, "max": 0xFF, "reserved": ()},
    "STORE_TEMP": {"store": True, "perm": False, "grfid": False, "max": 0x10F, "reserved": range(0x80, 0xFF + 1)},
    "LOAD_PERM": {"store": False, "perm": True, "grfid": True, "max": 0xFF, "reserved": ()},
    "LOAD_TEMP": {"store": False, "perm": False, "grfid": False, "max": 0xFF, "reserved": range(0x80, 0xFF + 1)},
}


class StorageOp(Expression):
    """
    Class for reading/writing to (temporary or permanent) storage

    @ivar name: Name of the called storage function
    @type name: C{str}

    @ivar info: Dictionary containing information about the operation to perform
    @type info: C{dict}

    @ivar value: Value to store, or C{None} for loading operations
    @type value: L{Expression}

    @ivar register: Register to access
    @type register: L{Expression}

    @ivar grfid: GRFID of the register to access
    @type grfid: L{Expression}

    """

    def __init__(self, name, args, pos=None):
        Expression.__init__(self, pos)
        self.name = name
        assert name in storage_op_info
        self.info = storage_op_info[name]

        arg_len = (2,) if self.info["store"] else (1,)
        if self.info["grfid"]:
            arg_len += (arg_len[0] + 1,)
        if len(args) not in arg_len:
            argstr = "{:d}".format(arg_len[0]) if len(arg_len) == 1 else "{}..{}".format(arg_len[0], arg_len[1])
            raise generic.ScriptError(
                "{} requires {} argument(s), encountered {:d}".format(name, argstr, len(args)), pos
            )

        i = 0
        if self.info["store"]:
            self.value = args[i]
            i += 1
        else:
            self.value = None

        self.register = args[i]
        i += 1

        if i < len(args):
            self.grfid = args[i]
            assert self.info["grfid"]
        else:
            self.grfid = None

    def debug_print(self, indentation):
        generic.print_dbg(indentation, self.name)
        generic.print_dbg(indentation + 2, "Register:")
        self.register.debug_print(indentation + 4)
        if self.value is not None:
            generic.print_dbg(indentation + 2, "Value:")
            self.value.debug_print(indentation + 4)
        if self.grfid is not None:
            generic.print_dbg(indentation + 2, "GRFID:")
            self.grfid.debug_print(indentation + 4)

    def __str__(self):
        args = []
        if self.value is not None:
            args.append(str(self.value))
        args.append(str(self.register))
        if self.grfid is not None:
            args.append(str(self.grfid))
        return "{}({})".format(self.name, ", ".join(args))

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        args = []
        if self.value is not None:
            value = self.value.reduce(id_dicts)
            if value.type() != Type.INTEGER:
                if value.type() == Type.SPRITEGROUP_REF:
                    raise generic.ProcCallSyntaxError(value.name, value.pos)
                raise generic.ScriptError("Value to store must be an integer.", value.pos)
            args.append(value)

        register = self.register.reduce(id_dicts)
        if register.type() != Type.INTEGER:
            if register.type() == Type.SPRITEGROUP_REF:
                raise generic.ProcCallSyntaxError(register.name, register.pos)
            raise generic.ScriptError("Register to access must be an integer.", register.pos)
        if isinstance(register, ConstantNumeric) and register.value > self.info["max"]:
            raise generic.ScriptError("Maximum register for {} is {:d}".format(self.name, self.info["max"]), self.pos)
        if isinstance(register, ConstantNumeric) and register.value in self.info["reserved"]:
            raise generic.ScriptError(
                "Temporary registers from 128 to 255 are reserved for NML's internal calculations.", self.pos
            )
        args.append(register)

        if self.grfid is not None:
            grfid = self.grfid.reduce(id_dicts)
            # Test validity
            parse_string_to_dword(grfid)
            args.append(grfid)

        return StorageOp(self.name, args, self.pos)

    def supported_by_action2(self, raise_error):
        return True

    def supported_by_actionD(self, raise_error):
        if raise_error:
            raise generic.ScriptError("{}() may only be used inside switch-blocks".format(self.name), self.pos)
        return False

    def collect_references(self):
        return self.register.collect_references() + (self.value.collect_references() if self.value is not None else [])

    def is_read_only(self):
        assert self.info["store"] == (self.value is not None)
        return (not self.info["store"]) and self.register.is_read_only()



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic


class FreeNumberList:
    """
    Contains a list with numbers and functions to pop one number from the list,
    to save the current state and to restore to the previous state.

    @ivar total_amount: Amount of numbers in the beginning.
    @type total_amount: C{int}

    @ivar stats: Statistics about id usage.
    @type stats: Tuple of C{int} and L{Position} referring to the amount on location of most concurrently used ids.

    @ivar free_numbers: The list with currently unused numbers.
    @type free_numbers: C{list}

    @ivar states: A list of lists. Each sublist contains all numbers
        that were L{popped<pop>} between the last call to L{save} and
        the next call to L{save}. Every time L{save} is called one
        sublist is added, every time L{restore} is called the topmost
        sublist is removed and it's values are added to the free number
        list again.
    @type states: C{list}

    @ivar used_numbers: A set with all numbers that have been used at
        some time. This is used by L{pop_unique}.
    @type used_numbers: C{set}

    @ivar exception: Exception to be thrown when there is no number available
        when requested via pop() or pop_global().
    @type exception: C{str}

    @ivar exception_unique: Exception to be thrown when there is no unique number
        available when it is requested via pop_unique().
    @type exception_unique: C{str}
    """

    def __init__(self, free_numbers, exception, exception_unique):
        self.total_amount = len(free_numbers)
        self.stats = (0, None)
        self.free_numbers = free_numbers
        self.states = []
        self.used_numbers = set()
        self.exception = exception
        self.exception_unique = exception_unique

    def pop(self, pos):
        """
        Pop a free number from the list. You have to call L{save} at least
        once before calling L{pop}.

        @param pos: Positional context.
        @type  pos: L{Position}

        @return: Some free number.
        """
        assert len(self.states) > 0
        if len(self.free_numbers) == 0:
            raise generic.ScriptError(self.exception, pos)
        num = self.free_numbers.pop()
        self.states[-1].append(num)
        self.used_numbers.add(num)

        num_used = len(self.used_numbers)
        if num_used > self.stats[0]:
            self.stats = (num_used, pos)

        return num

    def pop_global(self, pos):
        """
        Pop a free number from the list. The number may have been used before
        and already been restored, but it'll never be given out again.

        @param pos: Positional context.
        @type  pos: L{Position}

        @return: Some free number.
        """
        if len(self.free_numbers) == 0:
            raise generic.ScriptError(self.exception, pos)
        return self.free_numbers.pop()

    def pop_unique(self, pos):
        """
        Pop a free number from the list. The number has not been used before
        and will not be used again.

        @param pos: Positional context.
        @type  pos: L{Position}

        @return: A unique free number.
        """
        for num in reversed(self.free_numbers):
            if num in self.used_numbers:
                continue
            self.free_numbers.remove(num)
            self.used_numbers.add(num)

            num_used = len(self.used_numbers)
            if num_used > self.stats[0]:
                self.stats = (num_used, pos)

            return num
        raise generic.ScriptError(self.exception_unique, pos)

    def save(self):
        """
        Save the current state. All calls to L{pop} will be saved and can be
        reverted by calling L{restore}
        """
        self.states.append([])

    def restore(self):
        """
        Add all numbers given out via L{pop} since the last L{save} to the free
        number list.
        """
        assert len(self.states) > 0
        self.states[-1].reverse()
        self.free_numbers.extend(self.states[-1])
        self.states.pop()



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

try:
    from PIL import Image
except ImportError:
    # Image is required only when using graphics
    pass

from nml import expression, generic
from nml.actions import base_action
from nml.ast import assignment

FLAG_NOCROP = 0x0040
FLAG_NOALPHA = 0x0100
FLAG_WHITE = 0x0200
FLAG_ANIM = 0x0400

real_sprite_flags = {
    "CROP": 0,  # Allow cropping
    "NOCROP": FLAG_NOCROP,  # Disallow cropping
    "ALPHA": 0,  # Allow semi-transparency
    "NOALPHA": FLAG_NOALPHA,  # Warn about semi-transparency
    "WHITE": FLAG_WHITE,  # Allow pure-white
    "NOWHITE": 0,  # Warn about pure-white
    "ANIM": FLAG_ANIM,  # Allow anim colours
    "NOANIM": 0,  # Warn about anim colours
}

# fmt: off
palmap_d2w = [
      0, 215, 216, 136,  88, 106,  32,  33,  # 0..7
     40, 245,  10,  11,  12,  13,  14,  15,  # 8..15
     16,  17,  18,  19,  20,  21,  22,  23,  # 16..23
     24,  25,  26,  27,  28,  29,  30,  31,  # 24..31
     53,  54,  34,  35,  36,  37,  38,  39,  # 32..39
    178,  41,  42,  43,  44,  45,  46,  47,  # 40..47
     48,  49,  50,  51,  52,  53,  54,  55,  # 48..55
     56,  57,  58,  59,  60,  61,  62,  63,  # 56..63
     64,  65,  66,  67,  68,  69,  70,  71,  # 64..71
     72,  73,  74,  75,  76,  77,  78,  79,  # 72..79
     80,  81,  82,  83,  84,  85,  86,  87,  # 80..87
     96,  89,  90,  91,  92,  93,  94,  95,  # 88..95
     96,  97,  98,  99, 100, 101, 102, 103,  # 96..103
    104, 105,  53, 107, 108, 109, 110, 111,  # 104..111
    112, 113, 114, 115, 116, 117, 118, 119,  # 112..119
    120, 121, 122, 123, 124, 125, 126, 127,  # 120..127
    128, 129, 130, 131, 132, 133, 134, 135,  # 128..135
    170, 137, 138, 139, 140, 141, 142, 143,  # 136..143
    144, 145, 146, 147, 148, 149, 150, 151,  # 144..151
    152, 153, 154, 155, 156, 157, 158, 159,  # 152..159
    160, 161, 162, 163, 164, 165, 166, 167,  # 160..167
    168, 169, 170, 171, 172, 173, 174, 175,  # 168..175
    176, 177, 178, 179, 180, 181, 182, 183,  # 176..183
    184, 185, 186, 187, 188, 189, 190, 191,  # 184..191
    192, 193, 194, 195, 196, 197, 198, 199,  # 192..199
    200, 201, 202, 203, 204, 205, 206, 207,  # 200..207
    208, 209, 210, 211, 212, 213, 214, 215,  # 208..215
    216, 217, 246, 247, 248, 249, 250, 251,  # 216..223
    252, 253, 254, 227, 228, 229, 230, 231,  # 224..231
    232, 233, 234, 235, 236, 237, 238, 239,  # 232..239
    240, 241, 242, 243, 244, 217, 218, 219,  # 240..247
    220, 221, 222, 223, 224, 225, 226, 255,  # 248..255
]

palmap_w2d = [
      0,   1,   2,   3,   4,   5,   6,   7,  # 0..7
      8,   9,  10,  11,  12,  13,  14,  15,  # 8..15
     16,  17,  18,  19,  20,  21,  22,  23,  # 16..23
     24,  25,  26,  27,  28,  29,  30,  31,  # 24..31
      6,   7,  34,  35,  36,  37,  38,  39,  # 32..39
      8,  41,  42,  43,  44,  45,  46,  47,  # 40..47
     48,  49,  50,  51,  52,  53,  54,  55,  # 48..55
     56,  57,  58,  59,  60,  61,  62,  63,  # 56..63
     64,  65,  66,  67,  68,  69,  70,  71,  # 64..71
     72,  73,  74,  75,  76,  77,  78,  79,  # 72..79
     80,  81,  82,  83,  84,  85,  86,  87,  # 80..87
      4,  89,  90,  91,  92,  93,  94,  95,  # 88..95
     96,  97,  98,  99, 100, 101, 102, 103,  # 96..103
    104, 105,   5, 107, 108, 109, 110, 111,  # 104..111
    112, 113, 114, 115, 116, 117, 118, 119,  # 112..119
    120, 121, 122, 123, 124, 125, 126, 127,  # 120..127
    128, 129, 130, 131, 132, 133, 134, 135,  # 128..135
      3, 137, 138, 139, 140, 141, 142, 143,  # 136..143
    144, 145, 146, 147, 148, 149, 150, 151,  # 144..151
    152, 153, 154, 155, 156, 157, 158, 159,  # 152..159
    160, 161, 162, 163, 164, 165, 166, 167,  # 160..167
    168, 169, 170, 171, 172, 173, 174, 175,  # 168..175
    176, 177, 178, 179, 180, 181, 182, 183,  # 176..183
    184, 185, 186, 187, 188, 189, 190, 191,  # 184..191
    192, 193, 194, 195, 196, 197, 198, 199,  # 192..199
    200, 201, 202, 203, 204, 205, 206, 207,  # 200..207
    208, 209, 210, 211, 212, 213, 214,   1,  # 208..215
      2, 245, 246, 247, 248, 249, 250, 251,  # 216..223
    252, 253, 254, 229, 230, 231, 227, 228,  # 224..231
    232, 233, 234, 235, 236, 237, 238, 239,  # 232..239
    240, 241, 242, 243, 244,   9, 218, 219,  # 240..247
    220, 221, 222, 223, 224, 225, 226, 255,  # 248..255
]
# fmt: on

translate_w2d = bytearray(v for v in palmap_w2d)
translate_d2w = bytearray(v for v in palmap_d2w)


def convert_palette(pal):
    ret = 256 * [0]
    for idx, colour in enumerate(pal):
        if 0xD7 <= idx <= 0xE2:
            if idx != colour:
                raise generic.ScriptError(
                    "Indices 0xD7..0xE2 are not allowed in recolour sprites when the output is in the WIN palette"
                )
            continue
        ret[palmap_d2w[idx]] = palmap_d2w[colour]
    return ret


class RealSprite:
    """
    @ivar param_list: Original parameters from NML source file.
    @type param_list: List of L{expression.Expression}, or C{None}

    @ivar label: Optional label from NML source file.
    @type label: L{expression.Identifier} or C{None}

    @ivar is_empty: True, if this is an empty sprite with zero dimension.
    @type is_empty: C{bool}

    @ivar file: Filename of primary image file.
    @type file: L{expression.StringLiteral}

    @ivar xpos: X position of sprite in source image file.
    @type xpos: L{expression.ConstantNumeric} or C{None}

    @ivar ypos: Y position of sprite in source image file.
    @type ypos: L{expression.ConstantNumeric} or C{None}

    @ivar mask_file: Filename of additional mask image file.
    @type mask_file: L{expression.StringLiteral} or C{None}

    @ivar mask_pos: Position of sprite in source mask image file.
    @type mask_pos: Pair of L{expression.ConstantNumeric}, or C{None}

    @ivar xsize: X size of sprite in both source image and mask file.
    @type xsize: L{expression.ConstantNumeric} or C{None}

    @ivar ysize: Y position of sprite in both source image and mask file.
    @type ysize: L{expression.ConstantNumeric} or C{None}

    @ivar xrel: X sprite offset.
    @type xrel: L{expression.ConstantNumeric}

    @ivar yrel: Y sprite offset.
    @type yrel: L{expression.ConstantNumeric}

    @ivar flags: Cropping/warning flags.
    @type flags: L{expression.ConstantNumeric}

    @ivar poslist: Position of creation of the sprite, if available.
    @type poslist: C{list} of L{Position}
    """

    def __init__(self, param_list=None, label=None, poslist=None):
        self.param_list = param_list
        self.label = label
        self.is_empty = False
        self.xpos = None
        self.ypos = None
        self.xsize = None
        self.ysize = None
        if poslist is None:
            self.poslist = []
        else:
            self.poslist = poslist

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Real sprite, parameters:")
        for param in self.param_list:
            param.debug_print(indentation + 2)

    def get_labels(self):
        labels = {}
        if self.label is not None:
            labels[self.label.value] = 0
        return labels, 1

    def expand(self, default_file, default_mask_file, poslist, id_dict):
        return [parse_real_sprite(self, default_file, default_mask_file, poslist, id_dict)]

    def check_sprite_size(self):
        generic.check_range(self.xpos.value, 0, 0x7FFFFFFF, "Real sprite parameter 'xpos'", self.xpos.pos)
        generic.check_range(self.ypos.value, 0, 0x7FFFFFFF, "Real sprite parameter 'ypos'", self.ypos.pos)
        generic.check_range(self.xsize.value, 1, 0xFFFF, "Real sprite parameter 'xsize'", self.xsize.pos)
        generic.check_range(self.ysize.value, 1, 0xFFFF, "Real sprite parameter 'ysize'", self.ysize.pos)

    def validate_size(self):
        """
        Check if xpos/ypos/xsize/ysize are already set and if not, set them
        to 0,0,image_width,image_height.
        """
        if self.xpos is None:
            with Image.open(generic.find_file(self.file.value)) as im:
                self.xpos = expression.ConstantNumeric(0)
                self.ypos = expression.ConstantNumeric(0)
                self.xsize = expression.ConstantNumeric(im.size[0])
                self.ysize = expression.ConstantNumeric(im.size[1])
                self.check_sprite_size()
        if self.mask_pos is None:
            self.mask_pos = (self.xpos, self.ypos)

    def __str__(self):
        ret = ""
        if self.label is not None:
            ret += str(self.label) + ": "
        ret += "["
        ret += ", ".join([str(param) for param in self.param_list])
        ret += "]"
        return ret

    def get_cache_key(self, crop_sprites):
        """
        Assemble the sprite metadata into a key, able to identify the sprite.

        @param crop_sprites: Whether to crop sprites, which allow it.
        @type  crop_sprites: C{bool}

        @return: Key
        @rtype: C{tuple}
        """
        filename_8bpp = None
        filename_32bpp = None
        if self.bit_depth == 8:
            filename_8bpp = self.file
        else:
            filename_32bpp = self.file
            filename_8bpp = self.mask_file

        x = self.xpos.value
        y = self.ypos.value
        size_x = self.xsize.value
        size_y = self.ysize.value
        if self.bit_depth == 8 or self.mask_pos is None:
            mask_x, mask_y = x, y
        else:
            mask_x = self.mask_pos[0].value
            mask_y = self.mask_pos[1].value

        rgb_file, rgb_rect = (
            (filename_32bpp.value, (x, y, size_x, size_y)) if filename_32bpp is not None else (None, None)
        )
        mask_file, mask_rect = (
            (filename_8bpp.value, (mask_x, mask_y, size_x, size_y)) if filename_8bpp is not None else (None, None)
        )
        do_crop = crop_sprites and ((self.flags.value & FLAG_NOCROP) == 0)
        return (rgb_file, rgb_rect, mask_file, mask_rect, do_crop)


class SpriteAction(base_action.BaseAction):
    """
    @ivar sprite_num: Number of the sprite, or C{None} if not decided yet.
    @type sprite_num: C{int} or C{None}

    @ivar last: Whether this sprite action is the last of a series.
    @type last: C{bool}
    """

    def __init__(self):
        self.sprite_num = None
        self.last = False

    def prepare_output(self, sprite_num):
        if self.sprite_num is not None and self.sprite_num.value != sprite_num:
            msg = "Sprite number {:d} given in base_graphics-block, but it doesn't match output sprite number {:d}"
            msg = msg.format(self.sprite_num.value, sprite_num)
            raise generic.ScriptError(msg)


class RealSpriteAction(SpriteAction):
    def __init__(self):
        SpriteAction.__init__(self)
        self.sprite_list = []

    def add_sprite(self, sprite):
        self.sprite_list.append(sprite)

    def write(self, file):
        if len(self.sprite_list) == 0 or self.sprite_list[0].is_empty:
            file.print_empty_realsprite()
        else:
            file.print_sprite([s for s in self.sprite_list if not s.is_empty])
        if self.last:
            file.newline()


class RecolourSprite:
    def __init__(self, mapping, label=None, poslist=None):
        self.mapping = mapping
        self.label = label
        if poslist is None:
            self.poslist = []
        else:
            self.poslist = poslist

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Recolour sprite, mapping:")
        for recolour in self.mapping:
            generic.print_dbg(indentation + 2, "{}: {};".format(recolour.name, recolour.value))

    def get_labels(self):
        labels = {}
        if self.label is not None:
            labels[self.label.value] = 0
        return labels, 1

    def expand(self, default_file, default_mask_file, poslist, id_dict):
        # create new struct, needed for template expansion
        new_mapping = []
        for old_assignment in self.mapping:
            from_min_value = old_assignment.name.min.reduce_constant([id_dict])
            from_max_value = (
                from_min_value
                if old_assignment.name.max is None
                else old_assignment.name.max.reduce_constant([id_dict])
            )
            to_min_value = old_assignment.value.min.reduce_constant([id_dict])
            to_max_value = (
                None if old_assignment.value.max is None else old_assignment.value.max.reduce_constant([id_dict])
            )
            new_mapping.append(
                assignment.Assignment(
                    assignment.Range(from_min_value, from_max_value),
                    assignment.Range(to_min_value, to_max_value),
                    old_assignment.pos,
                )
            )
        return [RecolourSprite(new_mapping, poslist=poslist)]

    def __str__(self):
        ret = "" if self.label is None else str(self.label) + ": "
        ret += "recolour_sprite {\n"
        for recolour in self.mapping:
            ret += "{}: {};".format(recolour.name, recolour.value)
        ret += "}"
        return ret


class RecolourSpriteAction(SpriteAction):
    def __init__(self, sprite):
        SpriteAction.__init__(self)
        self.sprite = sprite
        self.output_table = []

    def prepare_output(self, sprite_num):
        SpriteAction.prepare_output(self, sprite_num)
        colour_mapping = {}
        for recolour in self.sprite.mapping:
            if (
                recolour.value.max is not None
                and recolour.name.max.value - recolour.name.min.value
                != recolour.value.max.value - recolour.value.min.value
            ):
                raise generic.ScriptError(
                    "From and to ranges in a recolour block need to have the same size", recolour.pos
                )
            for i in range(recolour.name.max.value - recolour.name.min.value + 1):
                idx = recolour.name.min.value + i
                val = recolour.value.min.value
                if recolour.value.max is not None:
                    val += i
                colour_mapping[idx] = val
        for i in range(256):
            if i in colour_mapping:
                colour = colour_mapping[i]
            else:
                colour = i
            self.output_table.append(colour)

    def write(self, file):
        file.start_sprite(257)
        file.print_bytex(0)
        if file.palette not in ("DEFAULT", "LEGACY"):
            raise generic.ScriptError(
                "Recolour sprites are only supported when writing to the DEFAULT (DOS) or LEGACY (WIN) palette."
                " If you don't have any real sprites use the commandline option -p to set a palette."
            )
        colour_table = self.output_table if file.palette == "DEFAULT" else convert_palette(self.output_table)
        for idx, colour in enumerate(colour_table):
            if idx % 16 == 0:
                file.newline()
            file.print_bytex(colour)
        if self.last:
            file.newline()
        file.end_sprite()


class TemplateUsage:
    def __init__(self, name, param_list, label, pos):
        self.name = name
        self.param_list = param_list
        self.label = label
        self.pos = pos

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Template used:", self.name.value)
        generic.print_dbg(indentation + 2, "Parameters:")
        for param in self.param_list:
            param.debug_print(indentation + 4)

    def get_labels(self):
        # Load labels from the template definition
        if self.name.value not in sprite_template_map:
            raise generic.ScriptError("Encountered unknown template identifier: " + self.name.value, self.name.pos)
        labels, offset = sprite_template_map[self.name.value].get_labels()
        # Add (possibly) label applied to ourselves
        if self.label is not None:
            if self.label.value in labels:
                raise generic.ScriptError(
                    "Duplicate label encountered; '{}' already exists.".format(self.label.value), self.pos
                )
            labels[self.label.value] = 0
        return labels, offset

    def expand(self, default_file, default_mask_file, poslist, parameters):
        if self.name.value not in sprite_template_map:
            raise generic.ScriptError("Encountered unknown template identifier: " + self.name.value, self.name.pos)
        template = sprite_template_map[self.name.value]
        if len(self.param_list) != len(template.param_list):
            raise generic.ScriptError(
                "Incorrect number of template arguments. Expected "
                + str(len(template.param_list))
                + ", got "
                + str(len(self.param_list)),
                self.pos,
            )
        param_dict = {}
        for i, param in enumerate(self.param_list):
            param = param.reduce([real_sprite_flags, parameters])
            if not isinstance(param, (expression.ConstantNumeric, expression.StringLiteral)):
                raise generic.ScriptError("Template parameters should be compile-time constants", param.pos)
            param_dict[template.param_list[i].value] = param.value

        return parse_sprite_list(
            template.sprite_list, default_file, default_mask_file, poslist + [self.pos], param_dict
        )

    def __str__(self):
        return "{}({})".format(self.name, ", ".join(str(param) for param in self.param_list))


def parse_real_sprite(sprite, default_file, default_mask_file, poslist, id_dict):
    # check the number of parameters
    num_param = len(sprite.param_list)
    if num_param == 0:
        sprite.is_empty = True
        return sprite
    elif not (2 <= num_param <= 9):
        raise generic.ScriptError(
            "Invalid number of arguments for real sprite. Expected 2..9.", sprite.param_list[0].pos
        )

    # create new sprite struct, needed for template expansion
    new_sprite = RealSprite(poslist=poslist + sprite.poslist)

    param_offset = 0

    if num_param >= 6:
        # xpos, ypos, xsize and ysize are all optional. If not specified they'll default
        # to 0, 0, image_width, image_height
        new_sprite.xpos = sprite.param_list[0].reduce_constant([id_dict])
        new_sprite.ypos = sprite.param_list[1].reduce_constant([id_dict])
        new_sprite.xsize = sprite.param_list[2].reduce_constant([id_dict])
        new_sprite.ysize = sprite.param_list[3].reduce_constant([id_dict])
        new_sprite.check_sprite_size()
        param_offset += 4

    new_sprite.xrel = sprite.param_list[param_offset].reduce_constant([id_dict])
    new_sprite.yrel = sprite.param_list[param_offset + 1].reduce_constant([id_dict])
    generic.check_range(
        new_sprite.xrel.value,
        -0x8000,
        0x7FFF,
        "Real sprite parameter {:d} 'xrel'".format(param_offset + 1),
        new_sprite.xrel.pos,
    )
    generic.check_range(
        new_sprite.yrel.value,
        -0x8000,
        0x7FFF,
        "Real sprite parameter {:d} 'yrel'".format(param_offset + 2),
        new_sprite.yrel.pos,
    )
    param_offset += 2

    # Next may follow any combination of (flags, filename, mask), but always in that order
    new_sprite.flags = expression.ConstantNumeric(0)
    if num_param > param_offset:
        try:
            new_sprite.flags = sprite.param_list[param_offset].reduce_constant([real_sprite_flags, id_dict])
            param_offset += 1
        except generic.ConstError:
            # No flags
            pass

    new_sprite.file = default_file
    if num_param > param_offset and not isinstance(sprite.param_list[param_offset], expression.Array):
        new_sprite.file = sprite.param_list[param_offset].reduce([id_dict])
        param_offset += 1
        if not isinstance(new_sprite.file, expression.StringLiteral):
            raise generic.ScriptError(
                "Real sprite parameter {:d} 'file' should be a string literal".format(param_offset + 1),
                new_sprite.file.pos,
            )

    if new_sprite.file is None:
        raise generic.ScriptError("No image file specified for real sprite", sprite.param_list[0].pos)

    new_sprite.mask_file = default_mask_file
    new_sprite.mask_pos = None
    if num_param > param_offset:
        mask = sprite.param_list[param_offset]
        param_offset += 1
        # Mask may be either string (file only)
        #   or array (empty => no mask, 1 value => file only, 2 => offsets only, 3 => file + offsets)
        if isinstance(mask, expression.Array):
            if not (0 <= len(mask.values) <= 3):
                raise generic.ScriptError(
                    "Real sprite mask should be an array with 0 to 3 values, encountered {:d}".format(len(mask.values)),
                    mask.pos,
                )
            if len(mask.values) == 0:
                # disable any default mask
                new_sprite.mask_file = None
            else:
                if len(mask.values) & 1:
                    new_sprite.mask_file = mask.values[0].reduce([id_dict])
                    if not isinstance(new_sprite.mask_file, expression.StringLiteral):
                        raise generic.ScriptError(
                            "Real sprite parameter 'mask_file' should be a string literal", new_sprite.file.pos
                        )
                if len(mask.values) & 2:
                    new_sprite.mask_pos = tuple(mask.values[i].reduce_constant([id_dict]) for i in range(-2, 0))
                    # Check that there is also a mask specified, else the offsets make no sense
                    if new_sprite.mask_file is None:
                        raise generic.ScriptError(
                            "Mask offsets are specified, but there is no mask file set.", new_sprite.mask_pos[0].pos
                        )
        else:
            new_sprite.mask_file = mask.reduce([id_dict])
            if not isinstance(new_sprite.mask_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "Real sprite parameter {:d} 'mask' should be an array or string literal".format(param_offset + 1),
                    new_sprite.file.pos,
                )

    if num_param > param_offset:
        raise generic.ScriptError(
            "Real sprite has too many parameters, the last {:d} parameter(s) cannot be parsed.".format(
                num_param - param_offset
            ),
            sprite.param_list[param_offset].pos,
        )

    return new_sprite


sprite_template_map = {}


def parse_sprite_list(sprite_list, default_file, default_mask_file, poslist, parameters=None):
    real_sprite_list = []
    for sprite in sprite_list:
        real_sprite_list.extend(sprite.expand(default_file, default_mask_file, poslist, parameters or {}))
    return real_sprite_list


def parse_sprite_data(sprite_container):
    """
    @param sprite_container: AST node that contains the sprite data
    @type sprite_container: L{SpriteContainer}

    @return: List of real sprite actions
    @rtype: C{list} of L{BaseAction}
    """
    all_sprite_data = sprite_container.get_all_sprite_data()
    action_list = []
    first = True

    for sprite_data in all_sprite_data:
        sprite_list, default_file, default_mask_file, pos, zoom_level, bit_depth = sprite_data
        new_sprite_list = parse_sprite_list(sprite_list, default_file, default_mask_file, [pos])
        if not first and len(new_sprite_list) != len(action_list):
            msg = "Expected {:d} alternative sprites for {} '{}', got {:d}."
            msg = msg.format(
                len(action_list), sprite_container.block_type, sprite_container.block_name.value, len(new_sprite_list)
            )
            raise generic.ScriptError(msg, sprite_container.pos)

        for i, sprite in enumerate(new_sprite_list):
            sprite.zoom_level = zoom_level
            sprite.bit_depth = bit_depth
            if (
                bit_depth == 8
                and isinstance(sprite, RealSprite)
                and (not sprite.is_empty)
                and sprite.mask_file is not None
            ):
                raise generic.ScriptError("Mask file may only be specified for 32bpp sprites.", sprite.mask_file.pos)
            if first:
                if isinstance(sprite, RealSprite):
                    action_list.append(RealSpriteAction())
                else:
                    assert isinstance(sprite, RecolourSprite)
                    action_list.append(RecolourSpriteAction(sprite))
            else:
                # Not the first sprite, so an alternative sprite
                if isinstance(sprite, RecolourSprite) or isinstance(action_list[i], RecolourSpriteAction):
                    raise generic.ScriptError(
                        "Alternative sprites may only be provided for and contain real sprites, not recolour sprites.",
                        sprite_container.pos,
                    )
                if action_list[i].sprite_list[0].is_empty and not sprite.is_empty:
                    # if the first sprite is empty, all others are ignored
                    generic.print_warning(
                        generic.Warning.OPTIMISATION,
                        "Alternative sprites for an empty real sprite are ignored.",
                        sprite_container.pos,
                    )
            if isinstance(sprite, RealSprite):
                action_list[i].add_sprite(sprite)
        first = False

    if len(action_list) != 0:
        action_list[-1].last = True
    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import re
import sys

from .ply import lex

from nml import expression, generic

# fmt: off
reserved = {
    "grf":                 "GRF",
    "var":                 "VARIABLE",
    "param":               "PARAMETER",
    "cargotable":          "CARGOTABLE",
    "railtypetable":       "RAILTYPETABLE",
    "roadtypetable":       "ROADTYPETABLE",
    "tramtypetable":       "TRAMTYPETABLE",
    "if":                  "IF",
    "else":                "ELSE",
    "while":               "WHILE",             # reserved
    "item":                "ITEM",              # action 0/3
    "property":            "PROPERTY",
    "graphics":            "GRAPHICS",
    "snowline":            "SNOWLINE",
    "basecost":            "BASECOST",
    "template":            "TEMPLATE",          # sprite template for action1
    "spriteset":           "SPRITESET",         # action 1
    "spritegroup":         "SPRITEGROUP",       # action 2
    "switch":              "SWITCH",            # deterministic varaction2
    "random_switch":       "RANDOMSWITCH",      # random action2
    "produce":             "PRODUCE",           # production action2
    "error":               "ERROR",             # action B
    "disable_item":        "DISABLE_ITEM",
    "replace":             "REPLACESPRITE",     # action A
    "replacenew":          "REPLACENEWSPRITE",  # action 5
    "font_glyph":          "FONTGLYPH",         # action 12
    "deactivate":          "DEACTIVATE",        # action E
    "town_names":          "TOWN_NAMES",        # action F
    "string":              "STRING",
    "return":              "RETURN",
    "livery_override":     "LIVERYOVERRIDE",
    "exit":                "SKIP_ALL",
    "tilelayout":          "TILELAYOUT",
    "spritelayout":        "SPRITELAYOUT",
    "alternative_sprites": "ALT_SPRITES",
    "base_graphics":       "BASE_GRAPHICS",
    "recolour_sprite":     "RECOLOUR_SPRITE",
    "engine_override":     "ENGINE_OVERRIDE",
    "sort":                "SORT_VEHICLES",
    "const":               "CONST"
}
# fmt: on

line_directive1_pat = re.compile(r'\#line\s+(\d+)\s*(\r?\n|"(.*)"\r?\n)')
line_directive2_pat = re.compile(r'\#\s+(\d+)\s+"(.*)"\s*((?:\d+\s*)*)\r?\n')


class NMLLexer:
    """
    @ivar lexer: PLY scanner object.
    @type lexer: L{ply.lex}

    @ivar includes: Stack of included files.
    @type includes: C{List} of L{generic.LinePosition}

    @ivar text: Input text to scan.
    @type text: C{str}
    """

    # Tokens
    tokens = list(reserved.values()) + [
        "ID",
        "PLUS",
        "MINUS",
        "TIMES",
        "DIVIDE",
        "MODULO",
        "AND",
        "OR",
        "XOR",
        "LOGICAL_AND",
        "LOGICAL_OR",
        "LOGICAL_NOT",
        "BINARY_NOT",
        "EQ",
        "LPAREN",
        "RPAREN",
        "SHIFT_LEFT",
        "SHIFT_RIGHT",
        "SHIFTU_RIGHT",
        "COMP_EQ",
        "COMP_NEQ",
        "COMP_LE",
        "COMP_GE",
        "COMP_LT",
        "COMP_GT",
        "COMMA",
        "RANGE",
        "LBRACKET",
        "RBRACKET",
        "LBRACE",
        "RBRACE",
        "TERNARY_OPEN",
        "COLON",
        "SEMICOLON",
        "STRING_LITERAL",
        "NUMBER",
        "FLOAT",
        "UNIT",
    ]

    t_PLUS = r"\+"
    t_MINUS = r"-"
    t_TIMES = r"\*"
    t_MODULO = r"%"
    t_DIVIDE = r"/"
    t_AND = r"&"
    t_OR = r"\|"
    t_XOR = r"\^"
    t_LOGICAL_AND = r"&&"
    t_LOGICAL_OR = r"\|\|"
    t_LOGICAL_NOT = r"!"
    t_BINARY_NOT = r"~"
    t_EQ = r"="
    t_LPAREN = r"\("
    t_RPAREN = r"\)"
    t_SHIFT_LEFT = r"<<"
    t_SHIFT_RIGHT = r">>"
    t_SHIFTU_RIGHT = r">>>"
    t_COMP_EQ = r"=="
    t_COMP_NEQ = r"!="
    t_COMP_LE = r"<="
    t_COMP_GE = r">="
    t_COMP_LT = r"<"
    t_COMP_GT = r">"
    t_COMMA = r","
    t_RANGE = r"\.\."
    t_LBRACKET = r"\["
    t_RBRACKET = r"\]"
    t_LBRACE = r"{"
    t_RBRACE = r"}"
    t_TERNARY_OPEN = r"\?"
    t_COLON = r":"
    t_SEMICOLON = r";"

    def t_FLOAT(self, t):
        r"\d+\.\d+"
        t.value = expression.ConstantFloat(float(t.value), t.lineno)
        return t

    def t_NUMBER(self, t):
        r"(0x[0-9a-fA-F]+)|(\d+)"
        base = 10
        if len(t.value) >= 2 and t.value[0:2] == "0x":
            t.value = t.value[2:]
            base = 16
        t.value = expression.ConstantNumeric(int(t.value, base), t.lineno)
        return t

    def t_UNIT(self, t):
        r"(nfo)|(mph)|(km/h)|(m/s)|(hpI)|(hpM)|(hp)|(kW)|(tons)|(ton)|(kg)|(snow%)"
        return t

    def t_ID(self, t):
        r"[a-zA-Z_][a-zA-Z0-9_]*"
        if t.value in reserved:  # Check for reserved words
            t.type = reserved[t.value]
        else:
            t.type = "ID"
            t.value = expression.Identifier(t.value, t.lineno)
        return t

    def t_STRING_LITERAL(self, t):
        r'"([^"\\]|\\.)*"'
        t.value = expression.StringLiteral(t.value[1:-1], t.lineno)
        return t

    # Ignored characters
    def t_ignore_comment(self, t):
        r"(/\*(\n|.)*?\*/)|(//.*)"
        self.increment_lines(t.value.count("\n"))

    def t_ignore_whitespace(self, t):
        r"[ \t\r]+"
        pass

    def t_line_directive1(self, t):
        r'\#line\s+\d+\s*(\r?\n|".*"\r?\n)'
        # See: https://gcc.gnu.org/onlinedocs/cpp/Line-Control.html
        m = line_directive1_pat.match(t.value)
        assert m is not None
        fname = self.lexer.lineno.filename if m.group(3) is None else m.group(3)

        # This type of line directive contains no information about includes, so we have to make some assumptions
        if self.includes and self.includes[-1].filename == fname:
            # Filename equal to the one on top of the stack -> end of an include
            self.includes.pop()
        elif fname != self.lexer.lineno.filename:
            # Not an end of include and not the current file -> start of an include
            self.includes.append(self.lexer.lineno)

        self.set_position(fname, int(m.group(1), 10))
        self.increment_lines(t.value.count("\n") - 1)

    def t_line_directive2(self, t):
        r'\#\s+\d+\s+".*"\s*(\d+\s*)*\r?\n'
        # Format: # lineno filename flags
        # See: https://gcc.gnu.org/onlinedocs/cpp/Preprocessor-Output.html
        m = line_directive2_pat.match(t.value)
        assert m is not None
        line, fname, flags = m.groups()
        line = int(line, 10)
        flags = [int(f, 10) for f in flags.split(" ") if f != ""] if flags is not None else []

        if 1 in flags:
            # File is being included, add current file/line to stack
            self.includes.append(self.lexer.lineno)
        elif 2 in flags:
            # End of include, new file should be equal to the one on top of the stack
            if self.includes and self.includes[-1].filename == fname:
                self.includes.pop()
            else:
                # But of course user input can never be trusted
                generic.print_warning(
                    generic.Warning.GENERIC,
                    "Information about included files is inconsistent, position information for errors may be wrong.",
                )

        self.set_position(fname, line)
        self.increment_lines(t.value.count("\n") - 1)

    def t_newline(self, t):
        r"\n+"
        self.increment_lines(len(t.value))

    def t_error(self, t):
        print(
            (
                "Illegal character '{}' (character code 0x{:02X}) at {}, column {:d}".format(
                    t.value[0], ord(t.value[0]), t.lexer.lineno, self.find_column(t)
                )
            )
        )
        sys.exit(1)

    def build(self):
        """
        Initial construction of the scanner.
        """
        self.lexer = lex.lex(module=self)

    def setup(self, text, fname):
        """
        Setup scanner for scanning an input file.

        @param text: Input text to scan.
        @type  text: C{str}

        @param fname: Filename associated with the input text (main input file).
        @type  fname: C{str}
        """
        self.includes = []
        self.text = text
        self.set_position(fname, 1)
        self.lexer.input(text)

    def set_position(self, fname, line):
        """
        @note: The lexer.lineno contains a Position object.
        """
        self.lexer.lineno = generic.LinePosition(fname, line, self.includes[:])

    def increment_lines(self, count):
        self.set_position(self.lexer.lineno.filename, self.lexer.lineno.line_start + count)

    def find_column(self, t):
        last_cr = self.text.rfind("\n", 0, t.lexpos)
        if last_cr < 0:
            last_cr = 0
        return t.lexpos - last_cr
